# -*- coding: utf-8 -*-
# wasp_general/crypto/random.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__


from Crypto.Random.random import getrandbits
import math

from wasp_general.verify import verify_type, verify_value


@verify_type(bits_count=int)
@verify_value(bits_count=lambda x: x >= 0)
def random_bits(bits_count):
	""" Random generator (PyCrypto getrandbits wrapper). The result is a non-negative value.

	:param bits_count: random bits to generate
	:return: int
	"""
	return getrandbits(bits_count)


@verify_type(maximum_value=int)
@verify_value(maximum_value=lambda x: x >= 0)
def random_int(maximum_value):
	""" Random generator (PyCrypto getrandbits wrapper). The result is a non-negative value.

	:param maximum_value: maximum integer value
	:return: int
	"""
	if maximum_value == 0:
		return 0
	elif maximum_value == 1:
		return random_bits(1)

	bits = math.floor(math.log2(maximum_value))
	result = random_bits(bits) + random_int(maximum_value - ((2 ** bits) - 1))
	return result


@verify_type(bytes_count=int)
@verify_value(bytes_count=lambda x: x >= 0)
def random_bytes(bytes_count):
	""" Generate random bytes sequence. (PyCrypto getrandbits wrapper)

	:param bytes_count: sequence length
	:return: bytes
	"""

	return bytes([getrandbits(8) for x in range(bytes_count)])



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/proto.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.


# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type


class WMessengerOnionProto(metaclass=ABCMeta):
	""" Abstract class for onion-messenger. Messengers job is divided into onions layers. Where each layer do its
	own small job. Layers are united in a session, that is used for message parsing or generation. Each layer
	has a name, which must be unique within single onion.

	Possible layer are transport encryption/decryption layers (rsa, aes,...), data encoding/decoding
	layers (base64, utf8,...), structure packing/unpacking layers (json, pickle, ...),
	lexical layers (shlex, json, pickle, ...), authentication/authorization layers and many more.
	"""

	@abstractmethod
	@verify_type(layer_name=str)
	def layer(self, layer_name):
		""" Return messengers layer by its name

		:param layer_name: name of a layer
		:return: WMessengerOnionLayerProto instance
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def layers_names(self):
		""" Available layers names

		:return: list of str
		"""
		raise NotImplementedError('This method is abstract')


class WMessengerEnvelopeProto(metaclass=ABCMeta):
	""" Each real processed message is wrapped in this class. This helps in object type checking (layers can
	check if message is a subclass of some envelope subclass) and helps to keep meta data, that may be
	generated/processed by layers
	"""

	@abstractmethod
	def message(self):
		""" Return real message. It can be anything - string, bytes, structure...

		:return: any-type object or None
		"""
		raise NotImplementedError('This method is abstract')

	def meta(self):
		""" Return message meta data (dictionary object). For dictionary keys, values usage scenario
		see current implementation (:class:`.WMessengerEnvelope`)

		:return: dict
		"""
		return {}


class WMessengerOnionSessionProto(metaclass=ABCMeta):
	""" Class represent messenger single session. Inside a onion messenger, this class process single message.
	"""

	@abstractmethod
	def onion(self):
		""" Return related onion. In most cases, it is the onion, that creates this session.
		:return: WMessengerOnionProto
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(envelope=(WMessengerEnvelopeProto, None))
	def process(self, envelope):
		""" Parse message, process it and generate response

		:param envelope: incoming or outgoing message or nothing. This value is passed to the first layer as is.
		:return: outgoing message or nothing. In most cases, this is a server response or client request.
		"""
		raise NotImplementedError('This method is abstract')


class WMessengerOnionLayerProto(metaclass=ABCMeta):
	""" Messenger layer, that do one simple job like encryption, encoding, parsing and etc.
	"""

	@verify_type(name=str)
	def __init__(self, name):
		""" Construct new layer with the given name

		:param name: name of the layer
		"""
		self.__name = name

	def name(self):
		""" Return the layer name

		:return: str
		"""
		return self.__name

	@abstractmethod
	@verify_type(envelope=WMessengerEnvelopeProto, session=WMessengerOnionSessionProto)
	def process(self, envelope, session, **kwargs):
		""" Parse/combine, decrypt/encrypt, decode/encode message.

		:param envelope: message to parse/combine/decrypt/encrypt/decode/encode
		:param session: related session
		:param kwargs: arguments that help to customize a layer (they are set in \
		:class:`.WMessengerOnionSessionFlowProto.IteratorInfo` objects)

		:return: WMessengerEnvelopeProto
		"""
		raise NotImplementedError('This method is abstract')


class WMessengerOnionSessionFlowProto(metaclass=ABCMeta):
	""" This class is used in the following class :class:`.WMessengerOnionSessionProto` to determine layer
	execution order.
	"""

	class IteratorInfo:
		""" Class that describes single layer call
		"""

		@verify_type(layer_name=str)
		def __init__(self, layer_name, **kwargs):
			""" Construct new descriptor

			:param layer_name: Layer name to be executed
			:param kwargs: layer arguments. see :meth:`.WMessengerOnionLayerProto.process` method
			"""
			self.__layer_name = layer_name
			self.__layer_kwargs = kwargs

		def layer_name(self):
			""" Return layer name

			:return: str
			"""
			return self.__layer_name

		def layer_args(self):
			""" Return layer arguments

			:return: dict
			"""
			return self.__layer_kwargs

	class Iterator(IteratorInfo):
		""" Iterator that is used to determine layers call sequence. Each iterator holds information for
		current layer call (:class:`.WMessengerOnionSessionFlowProto.IteratorInfo`) and layer to be called
		next. Iterators next layer (next iterator) can be defined at a runtime.
		"""

		@verify_type('paranoid', layer_name=str)
		def __init__(self, layer_name, next_iterator=None, **kwargs):
			""" Create iterator with the specified layer call information and the layer to be called next.

			:param layer_name: same as layer_name \
			in :meth:`WMessengerOnionSessionFlowProto.IteratorInfo.__init__` method
			:param next_iterator: For static execution order - next layer that should be called
			"""
			WMessengerOnionSessionFlowProto.IteratorInfo.__init__(self, layer_name, **kwargs)
			if next_iterator is not None:
				if isinstance(next_iterator, WMessengerOnionSessionFlowProto.IteratorInfo) is False:
					raise TypeError('Invalid type for next_iterator argument')
			self.__next_iterator = next_iterator

		@verify_type('paranoid', envelope=WMessengerEnvelopeProto)
		def next(self, envelope):
			""" Return next layer (iterator) to be called or None to stop execution

			:param envelope: message that was processed by a layer specified in this class
			:return: WMessengerOnionSessionFlowProto.Iterator or None
			"""
			return self.__next_iterator

	@abstractmethod
	@verify_type(envelope=WMessengerEnvelopeProto)
	def iterator(self, envelope):
		""" Return iterator to be used for message processing. Iterator may depend on incoming message

		:param envelope: original incoming message
		:return: WMessengerOnionSessionFlowProto.Iterator or None if there is no way to create session for the \
		given message
		"""
		raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/packers.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import json
from datetime import datetime

from wasp_general.verify import verify_type

from wasp_general.datetime import utc_datetime

from wasp_general.network.messenger.proto import WMessengerEnvelopeProto, WMessengerOnionSessionProto
from wasp_general.network.messenger.layers import WMessengerOnionPackerLayerProto
from wasp_general.network.messenger.envelope import WMessengerEnvelope, WMessengerTextEnvelope


class WMessengerJSONPacker(WMessengerOnionPackerLayerProto):

	__layer_name__ = "com.binblob.wasp-general.json-packer-layer"
	""" Layer name
	"""

	class JSONEncoder(json.encoder.JSONEncoder):

		def default(self, o):
			if isinstance(o, set) is True:
				return list(o)
			elif isinstance(o, bytes):
				return [x for x in o]
			elif isinstance(o, datetime):
				return utc_datetime(o, local_value=False).isoformat()
			return json.encoder.JSONEncoder.default(self, o)

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionPackerLayerProto.__init__(self, WMessengerJSONPacker.__layer_name__)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerEnvelopeProto)
	def pack(self, envelope, session, **kwargs):
		json_data = json.dumps(envelope.message(), cls=WMessengerJSONPacker.JSONEncoder)
		return WMessengerTextEnvelope(json_data, meta=envelope)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerTextEnvelope)
	def unpack(self, envelope, session, **kwargs):
		return WMessengerEnvelope(json.loads(envelope.message()), meta=envelope)



# -*- coding: utf-8 -*-
# wasp_general/task/scheduler/task_source.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from datetime import datetime, timedelta, MAXYEAR, timezone
from calendar import monthrange, weekday

from wasp_general.verify import verify_type, verify_value
from wasp_general.datetime import utc_datetime
from wasp_general.thread import WCriticalResource

from wasp_general.task.scheduler.proto import WTaskSourceProto, WScheduleRecord, WScheduleTask, WSchedulerServiceProto


# noinspection PyAbstractClass
class WBasicTaskSource(WTaskSourceProto):

	@verify_type(scheduler=WSchedulerServiceProto)
	def __init__(self, scheduler_service):
		self.__service = scheduler_service

	def scheduler_service(self):
		return self.__service


class WCronSchedule:

	__calendar__ = {1: 31, 2: 29, 3: 31, 4: 30, 5: 31, 6: 30, 7: 31, 8: 31, 9: 30, 10: 31, 11: 30, 12: 31}

	@verify_type(start_datetime=(datetime, None), minute=(int, None), hour=(int, None), day_of_month=(int, None))
	@verify_type(day_of_week=(int, None), month=(int, None))
	@verify_value(minute=lambda x: x is None or (0 <= x <= 59))
	@verify_value(hour=lambda x: x is None or (0 <= x <= 23))
	@verify_value(day_of_month=lambda x: x is None or (1 <= x <= 31))
	@verify_value(day_of_week=lambda x: x is None or (1 <= x <= 7))
	@verify_value(month=lambda x: x is None or (1 <= x <= 12))
	def __init__(self, start_datetime=None, minute=None, hour=None, day_of_month=None, day_of_week=None, month=None):

		if month is not None and day_of_month is not None and day_of_month > WCronSchedule.__calendar__[month]:
			raise ValueError('Invalid day and month combination')

		self.__start_datetime = start_datetime if start_datetime is not None else self.now()
		self.__next_start = None

		self.__minute = minute
		self.__hour = hour
		self.__day_of_month = day_of_month
		self.__day_of_week = day_of_week
		self.__month = month

		self.update()

	def start_datetime(self):
		return self.__start_datetime

	@verify_type(start_datetime=datetime)
	def _set_start_datetime(self, start_datetime):
		self.__start_datetime = start_datetime
		self.__next_start = None

	@verify_type('paranoid', year=int, month=int, day=int, hour=int, minute=int)
	@verify_value('paranoid', year=lambda x: x > 0)
	@verify_value('paranoid', month=lambda x: x is None or (1 <= x <= 12))
	@verify_value('paranoid', day=lambda x: x is None or (1 <= x <= 31))
	@verify_value('paranoid', hour=lambda x: x is None or (0 <= x <= 23))
	@verify_value('paranoid', minute=lambda x: x is None or (0 <= x <= 59))
	def _datetime(self, year, month, day, hour, minute):
		return datetime(year, month, day, hour, minute)

	def next_start(self):
		return self.__next_start

	def minute(self):
		return self.__minute

	def hour(self):
		return self.__hour

	def day_of_month(self):
		return self.__day_of_month

	def day_of_week(self):
		return self.__day_of_week

	def month(self):
		return self.__month

	def no_frequency(self):
		result = self.minute() is None and self.hour() is None and self.day_of_month() is None
		return result is True and self.day_of_week() is None and self.month() is None

	def final_datetime(self):
		return datetime(MAXYEAR, 12, 31, 23, 59)

	@classmethod
	def now(cls):
		return datetime.now()

	def update(self):
		if self.no_frequency() is True:
			return

		for (year, month) in self.month_iterator():
			for day in self.day_iterator(year, month):
				for (hour, minute) in self.time_iterator(year, month, day):
					self.__next_start = self._datetime(year, month, day, hour, minute)
					return

	@verify_type(omit_skipped=bool)
	def complete(self, omit_skipped=True):
		if self.no_frequency() is True:
			return

		next_start = self.next_start()
		seconds_last_minute = next_start.second
		seconds_last_hour = (next_start.minute * 60) + seconds_last_minute
		seconds_last_day = (next_start.hour * 24 * 60 * 60) + seconds_last_hour
		time_shift = 0
		if self.month() is not None:
			year = next_start.year
			month = next_start.month
			if month < 12:
				month += 1
			else:
				year += 1
				month = 1
			time_shift = (self._datetime(year, month, 1, 0, 0) - next_start).total_seconds()
		elif self.day_of_week() is not None:
			time_shift = (7 * 24 * 60 * 60) - seconds_last_day
		elif self.day_of_month() is not None:
			time_shift = (24 * 60 * 60) - seconds_last_day
		elif self.hour() is not None:
			time_shift = (60 * 60) - seconds_last_hour
		elif self.minute() is not None:
			time_shift = 60 - seconds_last_minute

		time_shift += 1
		new_start = next_start + timedelta(seconds=time_shift)
		if omit_skipped is True:
			now = self.now()
			if now > new_start:
				new_start = now

		self._set_start_datetime(new_start)
		self.update()

	def month_iterator(self):
		if self.no_frequency():
			return

		start_datetime = self.start_datetime()
		start_year = start_datetime.year
		start_month = start_datetime.month

		final_datetime = self.final_datetime()
		max_year = final_datetime.year
		max_month = final_datetime.month

		month = self.month()
		year = start_year

		if month is not None:
			if month < start_month:
				year += 1

			if year > max_year or (year == max_year and month > max_month):
				return

			while year < max_year or (year == max_year and month < max_month):
				yield ((year, month))
				year += 1

		else:  # month is None
			month = start_month
			while year < max_year or (year == max_year and month < max_month):
				yield ((year, month))

				if month < 12:
					month += 1
				else:
					year += 1
					month = 1

	def day_iterator(self, year, month):
		if self.no_frequency():
			return

		start_datetime = self.start_datetime()
		days_in_month = monthrange(year, month)[1]

		day_of_month = self.day_of_month()
		day_of_week = self.day_of_week()

		final_datetime = self.final_datetime()
		if year == final_datetime.year and month == final_datetime.month:
			days_in_month = final_datetime.day

		start_day = start_datetime.day
		if year != start_datetime.year or month != start_datetime.month:
			start_day = 1

		if day_of_month is None and day_of_week is None:
			day = start_day
			while day <= days_in_month:
				yield day
				day += 1
		elif day_of_month is not None and day_of_week is None:
			if start_day <= days_in_month:
				yield day_of_month
		elif day_of_month is not None and day_of_week is not None:
			if start_day <= days_in_month and weekday(year, month, day_of_month) == day_of_week:
				yield day_of_month
		else:  # day_of_month is None and day_of_week is not None
			current_weekday = weekday(year, month, start_day)

			delta = day_of_week - current_weekday
			if delta < 0:
				delta += 7
			day = start_day + delta

			while day <= days_in_month:
				yield day
				day += 7

	def time_iterator(self, year, month, day):
		start_datetime = self.start_datetime()
		max_hour = 23
		max_minute = 59

		hour = self.hour()
		minute = self.minute()

		final_datetime = self.final_datetime()
		if year == final_datetime.year and month == final_datetime.month and day == self.final_datetime():
			max_hour = final_datetime.hour
			max_minute = final_datetime.minute

		start_hour = start_datetime.hour
		start_minute = start_datetime.minute
		if year != start_datetime.year or month != start_datetime.month or day != start_datetime.day:
			start_hour = 0
			start_minute = 0

		if hour is None and minute is None:
			hour = start_hour
			minute = start_minute
			while start_hour <= hour < max_hour or (hour == max_hour and start_minute <= minute <= max_minute):
				yield ((hour, minute))
				if minute < 59:
					minute += 1
				else:
					minute = 0
					hour += 1
		elif hour is not None and minute is None:
			minute = start_minute
			if start_hour <= hour <= max_hour:
				while minute <= max_minute:
					yield ((hour, minute))
					minute += 1
		elif hour is not None and minute is not None:
			if start_hour <= hour < max_hour or (hour == max_hour and start_minute <= minute <= max_minute):
				yield ((hour, minute))
		else:  # hour is None and minute is not None
			hour = start_hour
			if start_minute > minute:
				hour += 1

			while start_hour <= hour < max_hour or (hour == max_hour and start_minute <= minute <= max_minute):
				yield ((hour, minute))
				hour += 1

	@classmethod
	@verify_type(scheduled=str)
	def from_string(cls, scheduled):
		return cls.from_string_tokens(*filter(lambda x: len(x) > 0, scheduled.strip().split(' ')))

	@classmethod
	@verify_type(tokens=str)
	def from_string_tokens(cls, *tokens):
		if len(tokens) != 5:
			raise ValueError('Malformed cron-schedule')

		tokens = [int(x) if x != '*' else None for x in tokens]

		if len(tokens) != 5:
			raise ValueError('Malformed cron-schedule')
		return cls(
			cls.now(), minute=tokens[0], hour=tokens[1], day_of_month=tokens[2], day_of_week=tokens[3],
			month=tokens[4]
		)


class WCronLocalTZSchedule(WCronSchedule):
	pass


class WCronUTCSchedule(WCronSchedule):

	@verify_type('paranoid', start_datetime=(datetime, None), minute=(int, None), hour=(int, None))
	@verify_type('paranoid', day_of_month=(int, None), day_of_week=(int, None), month=(int, None))
	@verify_value('paranoid', minute=lambda x: x is None or (0 <= x <= 59))
	@verify_value('paranoid', hour=lambda x: x is None or (0 <= x <= 23))
	@verify_value('paranoid', day_of_month=lambda x: x is None or (1 <= x <= 31))
	@verify_value('paranoid', day_of_week=lambda x: x is None or (1 <= x <= 7))
	@verify_value('paranoid', month=lambda x: x is None or (1 <= x <= 12))
	@verify_value(start_datetime=lambda x: x.tzinfo is not None and x.tzinfo == timezone.utc)
	def __init__(self, start_datetime=None, minute=None, hour=None, day_of_month=None, day_of_week=None, month=None):
		WCronSchedule.__init__(
			self, start_datetime, minute=minute, hour=hour, day_of_month=day_of_month,
			day_of_week=day_of_week, month=month
		)

	@verify_type(start_datetime=datetime)
	@verify_value(start_datetime=lambda x: x.tzinfo is not None and x.tzinfo == timezone.utc)
	def _set_start_datetime(self, start_datetime):
		WCronSchedule._set_start_datetime(self, start_datetime)

	@verify_type('paranoid', year=int, month=int, day=int, hour=int, minute=int)
	@verify_value('paranoid', year=lambda x: x > 0)
	@verify_value('paranoid', month=lambda x: x is None or (1 <= x <= 12))
	@verify_value('paranoid', day=lambda x: x is None or (1 <= x <= 31))
	@verify_value('paranoid', hour=lambda x: x is None or (0 <= x <= 23))
	@verify_value('paranoid', minute=lambda x: x is None or (0 <= x <= 59))
	def _datetime(self, year, month, day, hour, minute):
		return utc_datetime(datetime(year, month, day, hour, minute), local_value=False)

	@classmethod
	def now(cls):
		return utc_datetime()


class WCronScheduleRecord(WScheduleRecord):

	@verify_type('paranoid', task=WScheduleTask, task_group_id=(str, None))
	@verify_value('paranoid', on_drop=lambda x: x is None or callable(x))
	@verify_value('paranoid', on_wait=lambda x: x is None or callable(x))
	@verify_type(schedule=WCronSchedule, omit_skipped=bool)
	@verify_value(schedule=lambda x: isinstance(x, WCronLocalTZSchedule) or isinstance(x, WCronUTCSchedule))
	def __init__(
		self, cron_schedule, task, policy=None, task_group_id=None, on_drop=None, on_wait=None,
		omit_skipped=True
	):
		WScheduleRecord.__init__(
			self, task, policy=policy, task_group_id=task_group_id, on_drop=on_drop, on_wait=on_wait
		)
		self.__schedule = cron_schedule
		self.__omit_skipped = omit_skipped

	def cron_schedule(self):
		return self.__schedule

	def next_start(self):
		cron = self.cron_schedule()
		next_start = cron.next_start()
		if isinstance(cron, WCronLocalTZSchedule):
			return utc_datetime(dt=next_start)
		elif isinstance(cron, WCronUTCSchedule):
			return next_start
		raise RuntimeError('Corrupted object!')

	def complete(self):
		self.cron_schedule().complete(omit_skipped=self.__omit_skipped)


class WCronTaskSource(WBasicTaskSource, WCriticalResource):

	@verify_type('paranoid', scheduler_service=WSchedulerServiceProto)
	def __init__(self, scheduler_service):
		WBasicTaskSource.__init__(self, scheduler_service=scheduler_service)
		WCriticalResource.__init__(self)
		self.__tasks = []
		self.__next_task = None

	@verify_type(record=WCronScheduleRecord)
	def add_record(self, record):
		self.__add_record(record)
		self.scheduler_service().update(task_source=self)

	@WCriticalResource.critical_section()
	@verify_type('paranoid', record=WCronScheduleRecord)
	def __add_record(self, record):
		self.__tasks.append(record)
		self.__update(record)

	@verify_type('paranoid', task=(WCronScheduleRecord, None))
	def __update(self, task=None):
		if task is not None:
			next_start = task.next_start()
			if self.__next_task is None or next_start < self.__next_task.next_start():
				self.__next_task = task
		elif len(self.__tasks) > 0:
			next_task = self.__tasks[0]
			for task in self.__tasks[1:]:
				if next_task.next_start() is None:
					next_task = task
				elif task.next_start() is not None and task.next_start() < next_task.next_start():
					next_task = task

			self.__next_task = next_task
		else:
			self.__next_task = None

	@WCriticalResource.critical_section()
	def tasks_planned(self):
		return len(self.__tasks)

	@WCriticalResource.critical_section()
	def has_records(self):
		if self.__next_task is not None:
			next_start = self.__next_task.next_start()
			if next_start is not None and next_start <= utc_datetime():
				result = [self.__next_task]
				self.__next_task.complete()
				self.__update()
				return tuple(result)

	@WCriticalResource.critical_section()
	def next_start(self):
		if self.__next_task is not None:
			return self.__next_task.next_start()


class WInstantTaskSource(WBasicTaskSource, WCriticalResource):

	__lock_acquiring_timeout__ = 5
	""" Timeout with which critical section lock must be acquired
	"""

	@verify_type('paranoid', scheduler_service=WSchedulerServiceProto)
	def __init__(self, scheduler_service):
		WBasicTaskSource.__init__(self, scheduler_service=scheduler_service)
		WCriticalResource.__init__(self)
		self.__records = []

	@verify_type('paranoid', record=WScheduleRecord)
	def add_record(self, record):
		self.__add_record(record)
		self.scheduler_service().update(self)

	@verify_type(record=WScheduleRecord)
	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def __add_record(self, record):
		self.__records.append(record)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def has_records(self):
		if len(self.__records) > 0:
			result = self.__records.copy()
			self.__records = []
			return tuple(result)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def next_start(self):
		if len(self.__records) > 0:
			return utc_datetime()

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def tasks_planned(self):
		return len(self.__records)



# -*- coding: utf-8 -*-
# wasp_general/cli/curses.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
import curses

from wasp_general.verify import verify_type
from wasp_general.cli.cli import WConsoleWindowProto, WConsoleProto, WConsoleBase, WConsoleWindowBase
from wasp_general.cli.cli import WConsoleDrawerProto
from wasp_general.command.command import WCommandSet


class WCursesWindow(WConsoleWindowBase):

	class EmptyWindowDrawer(WConsoleDrawerProto):
		""" WConsoleWindowProto.DrawerProto implementation. Suites if there is nothing to display
		"""

		@verify_type('paranoid', window=WConsoleWindowProto, prompt_show=bool)
		def suitable(self, window, prompt_show=True):
			""" :meth:`WConsoleWindowProto.DrawerProto.suitable` method implementation
			"""
			if len(window.list_data(previous_data=True, console_row=True)) == 0:
				return True
			return False

		@verify_type('paranoid', window=WConsoleWindowProto, prompt_show=bool)
		def draw(self, window, prompt_show=True):
			""" :meth:`WConsoleWindowProto.DrawerProto.draw` method implementation
			"""
			window.set_cursor(0, 0)

	class SmallWindowDrawer(WConsoleDrawerProto):
		""" WConsoleWindowProto.DrawerProto implementation. Suites if there is content and content fits window
		width and height
		"""

		@verify_type('paranoid', window=WConsoleWindowProto, prompt_show=bool)
		def suitable(self, window, prompt_show=True):
			""" :meth:`WConsoleWindowProto.DrawerProto.suitable` method implementation
			"""
			if prompt_show is True:
				lines = len(window.list_data(previous_data=True, console_row=True))
			else:
				lines = len(window.list_data(previous_data=True))

			if 1 <= lines < (window.height() - 1):
				return True

			return False

		@verify_type('paranoid', window=WConsoleWindowProto, prompt_show=bool)
		def draw(self, window, prompt_show=True):
			""" :meth:`WConsoleWindowProto.DrawerProto.draw` method implementation
			"""
			if prompt_show is True:
				data_lines = window.list_data(previous_data=True, console_row=True)
			else:
				data_lines = window.list_data(previous_data=True)

			window.write_data(data_lines)

			if prompt_show is True:
				data_lines_to_cursor = window.list_data(
					previous_data=True, console_row_to_cursor=True
				)
				y = len(data_lines_to_cursor) - 1

				line_length = len(window.console().prompt()) + window.cursor()
				row_lines_to_cursor = window.list_data(console_row_to_cursor=True)
				line_length += (len(row_lines_to_cursor) - 1)  # append one char offset
				x = line_length % window.width()
			else:
				y = 0
				x = 0

			window.set_cursor(y, x)

	class ScrolledWindowDrawer(WConsoleDrawerProto):
		""" WConsoleWindowProto.DrawerProto implementation. Suites if content doesn't fit window width and
		height but current row fits
		"""

		@verify_type('paranoid', window=WConsoleWindowProto, prompt_show=bool)
		def suitable(self, window, prompt_show=True):
			""" :meth:`WConsoleWindowProto.DrawerProto.suitable` method implementation
			"""
			height = window.height()

			if prompt_show is True:
				lines = len(window.list_data(previous_data=True, console_row=True))
				console_row_lines = len(window.list_data(console_row=True))
				if (lines >= (height - 1)) and (console_row_lines < (height - 1)):
					return True
			else:
				lines = len(window.list_data(previous_data=True))
				if lines >= (height - 1):
					return True

			return False

		@verify_type('paranoid', window=WConsoleWindowProto, prompt_show=bool)
		def draw(self, window, prompt_show=True):
			""" :meth:`WConsoleWindowProto.DrawerProto.draw` method implementation
			"""
			height = window.height()
			if prompt_show is True:
				console_row_lines = window.list_data(console_row=True)
			else:
				console_row_lines = []

			delta = (height - (len(console_row_lines) + 1))
			previous_data = window.list_data(previous_data=True)
			delta_data = previous_data[(len(previous_data) - delta):]

			window.write_data(delta_data + console_row_lines)

			if prompt_show is True:
				lines_to_cursor = window.list_data(console_row_to_cursor=True)
				y = len(lines_to_cursor) - 1 + delta

				line_length = len(window.console().prompt()) + window.cursor()
				line_length += (len(lines_to_cursor) - 1)  # append one char offset
				x = line_length % window.width()
			else:
				y = 0
				x = 0

			window.set_cursor(y, x)

	class BigWindowDrawer(WConsoleDrawerProto):
		""" WConsoleWindowProto.DrawerProto implementation. Suites if content and even current row doesn't fit
		window width and height
		"""

		@verify_type('paranoid', window=WConsoleWindowProto, prompt_show=bool)
		def suitable(self, window, prompt_show=True):
			""" :meth:`WConsoleWindowProto.DrawerProto.suitable` method implementation
			"""
			if prompt_show is True:
				console_row_lines = len(window.list_data(console_row=True))
				if console_row_lines >= (window.height() - 1):
					return True
			return False

		@verify_type('paranoid', window=WConsoleWindowProto, prompt_show=bool)
		def draw(self, window, prompt_show=True):
			""" :meth:`WConsoleWindowProto.DrawerProto.draw` method implementation
			"""
			assert(prompt_show is True)

			height = window.height()
			lines = window.list_data(console_row=True)
			lines_to_cursor = window.list_data(console_row_to_cursor=True)
			lines_from_cursor = window.list_data(console_row_from_cursor=True)

			output_lines = []
			if len(lines_from_cursor) == 0:
				start = len(lines) - (height - 1)
				output_lines.extend(lines[start:])
				y = height - 2
			elif len(lines_from_cursor) < (height - 1):
				start = (len(lines)) - (height - 1) - (len(lines_from_cursor) - 1)
				output_lines.extend(lines[start:start + (height - 1)])
				y = (height - 1) - (len(lines) - len(lines_to_cursor)) - 1
			else:
				start = 0
				if len(lines_to_cursor) > 0:
					start = len(lines_to_cursor) - 1
				output_lines.extend(lines[start:(start + (height - 1))])
				y = 0

			window.write_data(output_lines)

			line_length = len(window.console().prompt()) + window.cursor()
			line_length += (len(lines_to_cursor) - 1)  # append one char offset
			x = line_length % window.width()
			window.set_cursor(y, x)

	@verify_type('paranoid', console=WConsoleProto)
	def __init__(self, console):
		WConsoleWindowBase.__init__(
			self, console, WCursesWindow.EmptyWindowDrawer(), WCursesWindow.SmallWindowDrawer(),
			WCursesWindow.ScrolledWindowDrawer(), WCursesWindow.BigWindowDrawer()
		)

	def width(self):
		return self.console().screen().getmaxyx()[1]

	def height(self):
		return self.console().screen().getmaxyx()[0]

	def clear(self):
		return self.console().screen().erase()

	def write_line(self, line_index, line):
		self.console().screen().addstr(line_index, 0, line)

	@verify_type('paranoid', prompt_show=bool)
	def refresh(self, prompt_show=True):
		WConsoleWindowBase.refresh(self, prompt_show=prompt_show)
		self.console().screen().refresh()

	def set_cursor(self, y, x):
		self.console().screen().move(y, x)


class WCursesKeyAction(metaclass=ABCMeta):

	def __init__(self, key):
		self.key = key

	@abstractmethod
	def action(self, console_meta):
		raise NotImplementedError('This method is abstract')

	def __call__(self, key, console_meta):
		if self.key == key:
			self.action(console_meta)


class WCursesKeyUp(WCursesKeyAction):

	def __init__(self):
		WCursesKeyAction.__init__(self, curses.KEY_UP)

	def action(self, console_meta):

		history = console_meta.history()
		if history.size() == 0:
			return

		if not console_meta.history_mode():
			console_meta.history_mode(True)
			history.position(history.size() - 1)
			console_meta.window().cursor(len(console_meta.row()))
		elif history.position() > 0:
			history.position(history.position() - 1)
			console_meta.window().cursor(len(console_meta.row()))
		else:
			return

		console_meta.refresh_window()


class WCursesKeyDown(WCursesKeyAction):

	def __init__(self):
		WCursesKeyAction.__init__(self, curses.KEY_DOWN)

	def action(self, console_meta):

		if not console_meta.history_mode():
			return

		history = console_meta.history()
		history_size = history.size()
		if history.position() < (history_size - 1):
			history.position(history.position() + 1)
			console_meta.window().cursor(len(console_meta.row()))
		else:
			console_meta.history_mode(False)
			console_meta.window().cursor(len(console_meta.row()))

		console_meta.refresh_window()


class WCursesKeyLeft(WCursesKeyAction):

	def __init__(self):
		WCursesKeyAction.__init__(self, curses.KEY_LEFT)

	def action(self, console_meta):
		cursor_position = console_meta.window().cursor()
		if cursor_position > 0 and len(console_meta.row()):
			cursor_position -= 1
			console_meta.window().cursor(cursor_position)
			console_meta.refresh_window()


class WCursesKeyRight(WCursesKeyAction):

	def __init__(self):
		WCursesKeyAction.__init__(self, curses.KEY_RIGHT)

	def action(self, console_meta):
		cursor_position = console_meta.window().cursor()
		if cursor_position < len(console_meta.row()):
			cursor_position += 1
			console_meta.window().cursor(cursor_position)
			console_meta.refresh_window()


class WCursesKeyBackspace(WCursesKeyAction):

	def __init__(self):
		WCursesKeyAction.__init__(self, curses.KEY_BACKSPACE)

	def action(self, console_meta):
		position = console_meta.window().cursor()

		command = console_meta.row()
		if 0 < position <= len(command):
			command = command[:(position - 1)] + command[position:]
			console_meta.window().cursor(position - 1)
			console_meta.update_row(command)
		console_meta.refresh_window()


class WCursesKeyDelete(WCursesKeyAction):

	def __init__(self):
		WCursesKeyAction.__init__(self, curses.KEY_DC)

	def action(self, console_meta):
		position = console_meta.window().cursor()
		command = console_meta.row()
		if 0 <= position < len(command):
			command = command[:position] + command[(position + 1):]
			console_meta.update_row(command)
		console_meta.refresh_window()


class WCursesKeyResize(WCursesKeyAction):

	def __init__(self):
		WCursesKeyAction.__init__(self, curses.KEY_RESIZE)

	def action(self, console_meta):
		console_meta.refresh_window()


class WCursesConsole(WConsoleBase):

	@verify_type('paranoid', command_set=(WCommandSet, None))
	def __init__(self, command_set=None):
		WConsoleBase.__init__(self, command_set=command_set)

		self.__screen = curses.initscr()
		self.__screen.keypad(True)
		self.__window = WCursesWindow(self)
		self.__key_actions = [
			WCursesKeyUp(), WCursesKeyDown(), WCursesKeyLeft(), WCursesKeyRight(), WCursesKeyBackspace(),
			WCursesKeyDelete(), WCursesKeyResize()
		]

		self.__stop_flag = True

	def window(self):
		return self.__window

	def screen(self):
		return self.__screen

	def stop(self):
		self.__stop_flag = True

	def start(self):

		self.__stop_flag = False
		while self.__stop_flag is False:
			self.start_session()

			while True:
				pressed_key = self.screen().get_wch()

				if isinstance(pressed_key, str):
					if pressed_key == '\n':
						break

					command = self.row()
					position = self.window().cursor()
					command = command[:position] + pressed_key + command[position:]
					self.update_row(command)
					self.window().cursor(position + 1)
					self.refresh_window()

				elif isinstance(pressed_key, int):
					for action in self.__key_actions:
						action(pressed_key, self)

			self.fin_session()

	def __del__(self):
		curses.endwin()



# -*- coding: utf-8 -*-
# wasp_general/task/scheduler/scheduler.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from datetime import timezone
from threading import Event

from wasp_general.verify import verify_type, verify_value, verify_subclass
from wasp_general.thread import WCriticalResource
from wasp_general.datetime import utc_datetime

from wasp_general.task.scheduler.proto import WScheduleTask, WRunningRecordRegistryProto, WSchedulerServiceProto
from wasp_general.task.scheduler.proto import WScheduleRecord, WTaskSourceProto
from wasp_general.task.thread import WPollingThreadTask


class WSchedulerWatchdog(WCriticalResource, WPollingThreadTask):
	""" Class that is looking for execution process of scheduled task. Each scheduled task has its own
	watchdog. Watchdog that will un-register stopped task from registry of running tasks
	"""

	__thread_polling_timeout__ = WPollingThreadTask.__thread_polling_timeout__ / 2
	""" Polling timeout
	"""

	__scheduled_task_startup_timeout__ = WPollingThreadTask.__thread_polling_timeout__ / 2
	""" Timeout for scheduled task to start (thread creation time)
	"""

	__lock_acquiring_timeout__ = 5
	""" Timeout with which critical section lock must be acquired
	"""

	__thread_name_prefix__ = 'TaskScheduler-Watchdog-'
	""" Prefix for the thread name (real thread name is made by concatenation of this prefix and schedule task uid)
	"""

	@classmethod
	@verify_type('paranoid', record=WScheduleRecord)
	def create(cls, record, registry):
		""" Core method for watchdog creation. Derived classes may redefine this method in order to change
		watchdog creation process

		:param record: schedule record that is ready to be executed
		:param registry: registry that is created this watchdog and registry that must be notified of \
		scheduled task stopping
		:return:
		"""
		return cls(record, registry)

	@verify_type(record=WScheduleRecord)
	def __init__(self, record, registry):
		""" Create new watch dog.

		:param record: schedule record that is ready to be executed
		:param registry: registry that is created this watch dog and registry that must be notified of \
		scheduled task stopping

		note: :class:`.WRunningRecordRegistry` is using :meth:`.WSchedulerWatchdog.create` method for watch
		dog creation
		"""
		WCriticalResource.__init__(self)
		WPollingThreadTask.__init__(self, thread_name=self.__thread_name_prefix__ + str(record.task_uid()))
		if isinstance(registry, WRunningRecordRegistry) is False:
			raise TypeError('Invalid registry type')

		self.__record = record
		self.__registry = registry
		self.__task = None

	def record(self):
		""" Return scheduler record

		:return: WScheduleRecord
		"""
		return self.__record

	def registry(self):
		""" Return parent registry

		:return: WRunningRecordRegistry
		"""
		return self.__registry

	def start(self):
		""" Start scheduled task and start watching

		:return: None
		"""
		self.__dog_started()
		WPollingThreadTask.start(self)

	def thread_started(self):
		""" Start watchdog thread function

		:return: None
		"""
		self.__thread_started()
		WPollingThreadTask.thread_started(self)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def __dog_started(self):
		""" Prepare watchdog for scheduled task starting

		:return: None
		"""
		if self.__task is not None:
			raise RuntimeError('Unable to start task. In order to start a new task - at first stop it')

		self.__task = self.record().task()
		if isinstance(self.__task, WScheduleTask) is False:
			task_class = self.__task.__class__.__qualname__
			raise RuntimeError('Unable to start unknown type of task: %s' % task_class)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def __thread_started(self):
		""" Start a scheduled task

		:return: None
		"""
		if self.__task is None:
			raise RuntimeError('Unable to start thread without "start" method call')
		self.__task.start()
		self.__task.start_event().wait(self.__scheduled_task_startup_timeout__)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def _polling_iteration(self):
		""" Poll for scheduled task stop events

		:return: None
		"""
		if self.__task is None:
			self.ready_event().set()
		elif self.__task.check_events() is True:
			self.ready_event().set()
			self.registry().task_finished(self)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def thread_stopped(self):
		""" Stop scheduled task beacuse of watchdog stop

		:return: None
		"""
		if self.__task is not None:
			if self.__task.stop_event().is_set() is False:
				self.__task.stop()
			self.__task = None


class WRunningRecordRegistry(WCriticalResource, WRunningRecordRegistryProto, WPollingThreadTask):
	""" Registry of started scheduled records. Has :meth:`.WRunningRecordRegistry.cleanup_event` event that is set
	when any of running scheduled task stopped. This event starts process of internal clean up (descriptors that
	were created for the record - will be removed)
	"""

	__thread_polling_timeout__ = WPollingThreadTask.__thread_polling_timeout__ / 4
	""" Polling timeout
	"""

	__watchdog_startup_timeout__ = WSchedulerWatchdog.__thread_polling_timeout__
	""" Timeout for watchdog to start (thread creation time)
	"""

	__lock_acquiring_timeout__ = 5
	""" Timeout with which critical section lock must be acquired
	"""

	__thread_name__ = 'SchedulerRegistry'
	""" Thread name (if thread_name_suffix was specified in constructor then suffix value is concatenated to
	this name)
	"""

	@verify_subclass(watchdog_cls=(WSchedulerWatchdog, None))
	@verify_type(thread_name_suffix=(str, None))
	def __init__(self, watchdog_cls=None, thread_name_suffix=None):
		""" Create new registry

		:param watchdog_cls: watchdog that should be used (:class:`.WSchedulerWatchdog` by default)
		:param thread_name_suffix: suffix to thread name
		"""
		WCriticalResource.__init__(self)
		WRunningRecordRegistryProto.__init__(self)
		thread_name = self.__thread_name__
		if thread_name_suffix is not None:
			thread_name += thread_name_suffix
		WPollingThreadTask.__init__(self, thread_name=thread_name)
		self.__running_registry = []
		self.__done_registry = []
		self.__cleanup_event = Event()
		self.__watchdog_cls = watchdog_cls if watchdog_cls is not None else WSchedulerWatchdog

	def cleanup_event(self):
		""" Return "cleanup" event

		:return: Event
		"""
		return self.__cleanup_event

	def watchdog_class(self):
		""" Return watchdog class that is used by this registry

		:return: WSchedulerWatchdog class or subclass
		"""
		return self.__watchdog_cls

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	@verify_type('paranoid', record=WScheduleRecord)
	def exec(self, record):
		""" Start the given schedule record (no time checks are made by this method, task will be started as is)

		:param record: schedule record to start

		:return: None
		"""
		watchdog_cls = self.watchdog_class()
		watchdog = watchdog_cls.create(record, self)
		watchdog.start()
		watchdog.start_event().wait(self.__watchdog_startup_timeout__)
		self.__running_registry.append(watchdog)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def running_records(self):
		""" Return schedule records that are running at the moment

		:return: tuple of WScheduleRecord
		"""
		return tuple([x.record() for x in self.__running_registry])

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def __len__(self):
		""" Return number of running records

		:return: int
		"""
		return len(self.__running_registry)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	@verify_type(watchdog=WSchedulerWatchdog)
	def task_finished(self, watchdog):
		""" Handle/process scheduled task stop

		:param watchdog: watchdog of task that was stopped

		:return: None
		"""
		if watchdog in self.__running_registry:  # when cleanup hits stop
			self.__running_registry.remove(watchdog)
			self.__done_registry.append(watchdog)
			self.cleanup_event().set()

	def _polling_iteration(self):
		""" Poll for cleanup event

		:return: None
		"""
		if self.cleanup_event().is_set() is True:
			self.cleanup()

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def cleanup(self):
		""" Do cleanup (stop and remove watchdogs that are no longer needed)

		:return: None
		"""
		for task in self.__done_registry:
			task.stop()
		self.__done_registry.clear()
		self.cleanup_event().clear()

	def thread_stopped(self):
		""" Handle registry stop

		:return: None
		"""
		self.cleanup()
		self.stop_running_tasks()

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def stop_running_tasks(self):
		""" Terminate all the running tasks

		:return: None
		"""
		for task in self.__running_registry:
			task.stop()
		self.__running_registry.clear()


class WPostponedRecordRegistry:
	""" Registry for postponed records.
	"""

	@verify_type(maximum_records=(int, None))
	@verify_value(maximum_records=lambda x: x is None or x >= 0)
	def __init__(self, maximum_records=None):
		""" Create new registry

		:param maximum_records: maximum number of tasks to postpone (no limit by default)
		"""
		self.__records = []
		self.__maximum_records = maximum_records

	def maximum_records(self):
		""" Return maximum number of records to postpone

		:return: int
		"""
		return self.__maximum_records

	@verify_type(record=WScheduleRecord)
	def postpone(self, record):
		""" Postpone (if required) the given task. The real action is depended on task postpone policy

		:param record: record to postpone
		:return: None
		"""

		maximum_records = self.maximum_records()
		if maximum_records is not None and len(self.__records) >= maximum_records:
			record.task_dropped()
			return

		task_policy = record.policy()
		task_group_id = record.task_group_id()

		if task_policy == WScheduleRecord.PostponePolicy.wait:
			self.__postpone_record(record)

		elif task_policy == WScheduleRecord.PostponePolicy.drop:
			record.task_dropped()

		elif task_policy == WScheduleRecord.PostponePolicy.postpone_first:
			if task_group_id is None:
				self.__postpone_record(record)
			else:
				record_found = None
				for previous_scheduled_record, task_index in self.__search_record(task_group_id):
					if previous_scheduled_record.policy() != task_policy:
						raise RuntimeError('Invalid tasks policies')
					record_found = previous_scheduled_record
					break

				if record_found is not None:
					record.task_dropped()
				else:
					self.__postpone_record(record)

		elif task_policy == WScheduleRecord.PostponePolicy.postpone_last:
			if task_group_id is None:
				self.__postpone_record(record)
			else:
				record_found = None
				for previous_scheduled_record, task_index in self.__search_record(task_group_id):
					if previous_scheduled_record.policy() != task_policy:
						raise RuntimeError('Invalid tasks policies')
					record_found = task_index
					break

				if record_found is not None:
					self.__records.pop(record_found).task_dropped()

				self.__postpone_record(record)
		else:
			raise RuntimeError('Invalid policy spotted')

	@verify_type(task_group_id=str)
	def __search_record(self, task_group_id):
		""" Search (iterate over) for tasks with the given task id

		:param task_group_id: target id

		:return: None
		"""
		for i in range(len(self.__records)):
			record = self.__records[i]
			if record.task_group_id() == task_group_id:
				yield record, i

	@verify_type(record=WScheduleRecord)
	def __postpone_record(self, record):
		""" Save the record and trigger 'postpone' method

		:param record: record to save

		:return: None
		"""
		self.__records.append(record)
		record.task_postponed()

	def has_records(self):
		""" Check if there are postponed records. True - there is at least one postpone record,
		False - otherwise

		:return: bool
		"""
		return len(self.__records) > 0

	def __len__(self):
		""" Return number of postponed records

		:return: int
		"""
		return len(self.__records)

	def __iter__(self):
		""" Iterate over postponed records. Once record is yield from this method, this record is removed
		from registry immediately

		:return: None
		"""
		while len(self.__records) > 0:
			yield self.__records.pop(0)


class WTaskSourceRegistry:
	""" Registry of tasks sources. It works as a dynamic queue - every task source notify this registry when
	next task should be started. And this registry fetches those tasks that is about to start. Registry is able
	to return schedule records from different sources at one time.
	"""

	def __init__(self):
		""" Create new registry
		"""
		self.__sources = {}

		self.__next_start = None
		self.__next_sources = []

	@verify_type(task_source=WTaskSourceProto)
	def add_source(self, task_source):
		""" Add new tasks source

		:param task_source:

		:return: None
		"""
		next_start = task_source.next_start()
		self.__sources[task_source] = next_start
		self.__update(task_source)

	@verify_type(task_source=(WTaskSourceProto, None))
	def update(self, task_source=None):
		""" Recheck next start of records from all the sources (or from the given one only)

		:param task_source: if defined - source to check

		:return: None
		"""
		if task_source is not None:
			self.__update(task_source)
		else:
			self.__update_all()

	def __update_all(self):
		""" Recheck next start of records from all the sources

		:return: None
		"""
		self.__next_start = None
		self.__next_sources = []

		for source in self.__sources:
			self.__update(source)

	@verify_type(task_source=WTaskSourceProto)
	def __update(self, task_source):
		""" Recheck next start of tasks from the given one only

		:param task_source: source to check

		:return: None
		"""
		next_start = task_source.next_start()
		if next_start is not None:

			if next_start.tzinfo is None or next_start.tzinfo != timezone.utc:
				raise ValueError('Invalid timezone information')

			if self.__next_start is None or next_start < self.__next_start:
				self.__next_start = next_start
				self.__next_sources = [task_source]
			elif next_start == self.__next_start:
				self.__next_sources.append(task_source)

	def check(self):
		""" Check if there are records that are ready to start and return them if there are any

		:return: tuple of WScheduleRecord or None (if there are no tasks to start)
		"""
		if self.__next_start is not None:
			utc_now = utc_datetime()
			if utc_now >= self.__next_start:
				result = []

				for task_source in self.__next_sources:
					records = task_source.has_records()
					if records is not None:
						result.extend(records)

				self.__update_all()

				if len(result) > 0:
					return tuple(result)

	def task_sources(self):
		""" Return task sources that was added to this registry

		:return: tuple of WTaskSourceProto
		"""
		return tuple(self.__sources.keys())


class WSchedulerService(WCriticalResource, WSchedulerServiceProto, WPollingThreadTask):
	""" Main scheduler service. This class unites different registries to present entire scheduler
	"""

	__thread_polling_timeout__ = WPollingThreadTask.__thread_polling_timeout__ / 8
	""" Polling timeout
	"""
	__lock_acquiring_timeout__ = 5
	""" Timeout with which critical section lock must be acquired
	"""

	__default_maximum_running_records__ = 10
	""" Number of records that are able to run simultaneously. This value is used by default
	"""

	__thread_name_prefix__ = 'TaskScheduler'
	""" Thread name (if thread_name_suffix was specified in constructor then suffix value is concatenated to
	this name)
	"""

	@verify_type('paranoid', maximum_postponed_records=(int, None))
	@verify_value('paranoid', maximum_postponed_records=lambda x: x is None or x > 0)
	@verify_type(maximum_running_records=(int, None), running_record_registry=(WRunningRecordRegistry, None))
	@verify_type(postponed_record_registry=(WPostponedRecordRegistry, None))
	@verify_type(thread_name_suffix=(str, None))
	@verify_value(maximum_running_records=lambda x: x is None or x > 0)
	def __init__(
		self, maximum_running_records=None, running_record_registry=None, maximum_postponed_records=None,
		postponed_record_registry=None, thread_name_suffix=None
	):
		""" Create new scheduler

		:param maximum_running_records: number of records that are able to run simultaneously \
		(WSchedulerService.__default_maximum_running_records__ is used as default value)
		:param running_record_registry: registry for running records
		:param maximum_postponed_records: number of records that are able to be postponed (no limit by default)
		:param postponed_record_registry: registry for postponed records
		:param thread_name_suffix: suffix to thread name
		"""
		WCriticalResource.__init__(self)
		WSchedulerServiceProto.__init__(self)
		thread_name = self.__thread_name_prefix__
		if thread_name_suffix is not None:
			thread_name += thread_name_suffix
		WPollingThreadTask.__init__(self, thread_name=thread_name)

		if maximum_postponed_records is not None and postponed_record_registry is not None:
			raise ValueError(
				'Conflict values found. Unable to instantiate scheduler service with '
				'"maximum_postponed_tasks" and "postponed_tasks_registry" values (chose one)'
			)

		default = lambda x, y: x if x is not None else y

		self.__maximum_running_records = default(
			maximum_running_records, self.__class__.__default_maximum_running_records__
		)

		self.__running_record_registry = default(running_record_registry, WRunningRecordRegistry())
		self.__postponed_record_registry = WPostponedRecordRegistry(maximum_postponed_records)
		self.__sources_registry = WTaskSourceRegistry()

		self.__awake_at = None

	def maximum_running_records(self):
		""" Return number of tasks that are able to run simultaneously

		:return: int
		"""
		return self.__maximum_running_records

	def maximum_postponed_records(self):
		""" Return number of tasks that are able to be postponed

		:return: int or None (for no limit)
		"""
		return self.__postponed_record_registry.maximum_records()

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	@verify_type('paranoid', task_source=WTaskSourceProto)
	def add_task_source(self, task_source):
		""" Add tasks source

		:param task_source: task source to add

		:return: None
		"""
		self.__sources_registry.add_source(task_source)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def task_sources(self):
		""" Return task sources that was added to this scheduler

		:return: tuple of WTaskSourceProto
		"""
		return self.__sources_registry.task_sources()

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	@verify_type('paranoid', task_source=(WTaskSourceProto, None))
	def update(self, task_source=None):
		""" Recheck next start of tasks from all the sources (or from the given one only)

		:param task_source: if defined - source to check

		:return: None
		"""
		self.__sources_registry.update(task_source=task_source)

	def running_records(self):
		""" Return scheduled tasks that are running at the moment

		:return: tuple of WScheduleRecord
		"""
		return self.__running_record_registry.running_records()

	def records_status(self):
		""" Return number of running and postponed tasks

		:return: tuple of two ints (first - running tasks, second - postponed tasks)
		"""
		return len(self.__running_record_registry), len(self.__postponed_record_registry)

	def thread_started(self):
		""" Start required registries and start this scheduler

		:return: None
		"""
		self.__running_record_registry.start()
		self.__running_record_registry.start_event().wait()
		WPollingThreadTask.thread_started(self)

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def _polling_iteration(self):
		""" Poll for different scheduler events like: there are tasks to run, there are tasks to postpone
		there are postponed tasks that should be running

		:return: None
		"""
		scheduled_tasks = self.__sources_registry.check()
		has_postponed_tasks = self.__postponed_record_registry.has_records()
		maximum_tasks = self.maximum_running_records()

		if scheduled_tasks is not None or has_postponed_tasks is not None:
			running_tasks = len(self.__running_record_registry)

			if running_tasks >= maximum_tasks:
				if scheduled_tasks is not None:
					for task in scheduled_tasks:
						self.__postponed_record_registry.postpone(task)
			else:
				if has_postponed_tasks is True:
					for postponed_task in self.__postponed_record_registry:
						self.__running_record_registry.exec(postponed_task)
						running_tasks += 1
						if running_tasks >= maximum_tasks:
							break

				if scheduled_tasks is not None:
					for task in scheduled_tasks:
						if running_tasks >= maximum_tasks:
							self.__postponed_record_registry.postpone(task)
						else:
							self.__running_record_registry.exec(task)
							running_tasks += 1

	@WCriticalResource.critical_section(timeout=__lock_acquiring_timeout__)
	def thread_stopped(self):
		""" Stop registries and this scheduler

		:return: None
		"""
		self.__running_record_registry.stop()



# -*- coding: utf-8 -*-
# wasp_general/network/web/headers.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import re
from io import StringIO
import email.message
from wasp_general.verify import verify_type, verify_value
from wasp_general.network.web.cookies import WHTTPCookieJar
from wasp_general.network.web.re_statements import http_header_name


class WHTTPHeaders:
	""" Represent HTTP Headers as they are described in RFC 1945. Call :meth:`.WHTTPHeaders.ro` method to create
	read-only copy (in this state no changes are allowed)

	Cookies that are stored with :meth:`.WHTTPHeaders.set_cookie_jar` method have highest priority than cookies
	stored within constructor and/or :meth:`.WHTTPHeaders.add_header`. So in the conflict situation cookies saved
	within :meth:`.WHTTPHeaders.set_cookie_jar` must be used.

	Cookies that are stored in :meth:`.WHTTPHeaders.client_cookie_jar` are client cookies (i.e. read-only
	cookies).
	"""

	header_name_re = re.compile(http_header_name)
	"""
	Regexp is for HTTP header name checking
	"""

	@staticmethod
	def header_name_check(header_name):
		""" Check header name for validity. Return True if name is valid

		:param header_name: name to check
		:return: bool
		"""
		header_match = WHTTPHeaders.header_name_re.match(header_name.encode('us-ascii'))
		return len(header_name) > 0 and header_match is not None

	def __init__(self, **kwargs):
		""" Construct new headers collection

		:param kwargs: dictionary of header names and corresponding values
		"""
		self.__headers = {}
		self.__ro_flag = False
		self.__normalization_mode = '1.0'
		self.__set_cookies = WHTTPCookieJar()
		for arg_name in kwargs.keys():
			self.add_headers(arg_name, kwargs[arg_name])

	def headers(self):
		""" Get specified header names

		:return: tuple of str
		"""
		return tuple(self.__headers.keys())

	@verify_type('paranoid', header_name=str)
	@verify_value('paranoid', header_name=lambda x: WHTTPHeaders.header_name_check(x))
	def remove_headers(self, header_name):
		""" Remove header by its name

		:param header_name: name of header to remove
		:return: None
		"""
		if self.__ro_flag:
			raise RuntimeError('ro')
		header_name = self.normalize_name(header_name)
		if header_name in self.__headers.keys():
			self.__headers.pop(header_name)

	@verify_type('paranoid', header_name=str)
	@verify_value('paranoid', header_name=lambda x: WHTTPHeaders.header_name_check(x))
	@verify_type(value=str, values=str)
	def add_headers(self, header_name, value, *values):
		""" Add new header

		:param header_name: name of the header to add
		:param value: header value
		:param values: additional header values (in a result request/response must be concatenated by the coma \
		or by the separate header string)
		:return: None
		"""
		if self.__ro_flag:
			raise RuntimeError('ro')
		header_name = self.normalize_name(header_name)
		if header_name not in self.__headers.keys():
			self.__headers[header_name] = [value]
		else:
			self.__headers[header_name].append(value)

		for single_value in values:
			self.__headers[header_name].append(single_value)

	@verify_type('paranoid', header_name=str, value=str, values=str)
	@verify_value('paranoid', header_name=lambda x: WHTTPHeaders.header_name_check(x))
	def replace_headers(self, header_name, value, *values):
		""" Replace header value with specified value and/or values

		:param header_name: target header
		:param value: new header value
		:param values: additional header values (in a result request/response must be concatenated by the coma \
		or by the separate header string)
		:return: None
		"""
		if self.__ro_flag:
			raise RuntimeError('ro')
		header_name = self.normalize_name(header_name)
		self.remove_headers(header_name)
		self.add_headers(header_name, value, *values)

	@verify_type('paranoid', header_name=str)
	@verify_value('paranoid', header_name=lambda x: WHTTPHeaders.header_name_check(x))
	def get_headers(self, header_name):
		""" Return header value by its name

		:param header_name: header name
		:return: tuple of str
		"""
		header_name = self.normalize_name(header_name)
		if header_name in self.__headers.keys():
			return tuple(self.__headers[header_name])

	@verify_type('paranoid', item=str)
	@verify_value('paranoid', item=lambda x: WHTTPHeaders.header_name_check(x))
	def __getitem__(self, item):
		""" Return header value by its name

		:param item: header name
		:return: tuple of str
		"""
		return self.get_headers(item)

	@verify_type(header_name=str)
	@verify_value(header_name=lambda x: WHTTPHeaders.header_name_check(x))
	def normalize_name(self, header_name):
		""" Return header name as it is recommended (required) by corresponding http protocol. For
		protocol switching use :meth:`.WHTTPHeaders.switch_name_style` method.

		All current available protocols (0.9-2) compare header names in a case-insensitive fashion. However,
		previous protocol versions (0.9-1.1) recommends to use camel-case names like Foo or Foo-Bar. But
		HTTP/2 (RFC 7540) strictly requires lowercase only header names.

		:param header_name: name to convert
		:return: str
		"""
		if self.__normalization_mode in ['0.9', '1.0', '1.1']:
			return '-'.join([x.capitalize() for x in header_name.split('-')])
		elif self.__normalization_mode == '2':
			return header_name.lower()
		raise RuntimeError('Internal error: unknown http protocol: %s' % self.__normalization_mode)

	@verify_type(http_protocol_version=str)
	@verify_value(http_protocol_version=lambda x: x in ['0.9', '1.0', '1.1', '2'])
	def switch_name_style(self, http_protocol_version):
		""" Return object copy with header names saved as it is described in the given protocol version

		see :meth:`.WHTTPHeaders.normalize_name`

		:param http_protocol_version: target HTTP protocol version
		:return: WHTTPHeaders
		"""
		new_headers = WHTTPHeaders()
		new_headers.__normalization_mode = http_protocol_version
		names = self.headers()
		for name in names:
			new_headers.add_headers(name, *self.get_headers(name))
		for cookie_name in self.__set_cookies.cookies():
			new_headers.__set_cookies.add_cookie(self.__set_cookies[cookie_name].copy())
		return new_headers

	def ro(self):
		""" Return read-only copy of this object

		:return: WHTTPHeaders
		"""
		ro_headers = WHTTPHeaders()
		names = self.headers()
		for name in names:
			ro_headers.add_headers(name, *self.get_headers(name))
		ro_headers.__cookies = self.__set_cookies.ro()
		ro_headers.__ro_flag = True
		return ro_headers

	@verify_type('paranoid', value=(str, None))
	def content_type(self, value=None):
		""" Set (replace) and or get "Content-Type" header value

		:param value: value to set (if specified)
		:return: None if header doesn't exist, otherwise - str
		"""
		content_type = self.normalize_name('Content-Type')
		if value is not None:
			self.replace_headers(content_type, value)
		if content_type in self.__headers.keys():
			return self.__headers[content_type][0]

	def set_cookie_jar(self):
		""" Return internal cookie jar that must be used for HTTP-response

		see :class:`.WHTTPCookieJar`

		:return: WHTTPCookieJar
		"""
		return self.__set_cookies

	def client_cookie_jar(self):
		""" Return internal cookie jar that must be used as HTTP-request cookies

		see :class:`.WHTTPCookieJar`

		:return: WHTTPCookieJar
		"""
		cookie_jar = WHTTPCookieJar()
		cookie_header = self.get_headers('Cookie')
		for cookie_string in (cookie_header if cookie_header is not None else tuple()):
			for single_cookie in WHTTPCookieJar.import_header_text(cookie_string):
				cookie_jar.add_cookie(single_cookie)

		return cookie_jar.ro()

	@classmethod
	@verify_type(http_code=str)
	def import_headers(cls, http_code):
		""" Create WHTTPHeaders by the given code. If code has 'Set-Cookie' headers, that headers are
		parsed, data are stored in internal cookie jar. At the end of parsing 'Set-Cookie' headers are
		removed from the result

		:param http_code: HTTP code to parse
		:return: WHTTPHeaders
		"""
		headers = WHTTPHeaders()
		message = email.message_from_file(StringIO(http_code))
		for header_name, header_value in message.items():
			headers.add_headers(header_name, header_value)

		cookie_header = headers.get_headers('Set-Cookie')
		if cookie_header is not None:
			for cookie_string in cookie_header:
				for single_cookie in WHTTPCookieJar.import_header_text(cookie_string):
					headers.set_cookie_jar().add_cookie(single_cookie)
			headers.remove_headers('Set-Cookie')

		return headers



# -*- coding: utf-8 -*-

import pytest

from wasp_general.cache import WCacheStorage, WGlobalSingletonCacheStorage, WInstanceSingletonCacheStorage
from wasp_general.cache import cache_control


class TestWCacheStorage:

	@staticmethod
	def foo():
		pass

	@staticmethod
	def bar():
		pass

	def test(self):

		assert(issubclass(WCacheStorage.CacheMissedException, Exception) is True)

		cache_entry = WCacheStorage.CacheEntry()
		assert(cache_entry.has_value is False)
		assert(cache_entry.cached_value is None)

		cache_entry = WCacheStorage.CacheEntry(has_value=True, cached_value=1)
		assert(cache_entry.has_value is True)
		assert(cache_entry.cached_value == 1)

		foo = TestWCacheStorage.foo

		pytest.raises(TypeError, WCacheStorage)
		pytest.raises(NotImplementedError, WCacheStorage.put, None, None, foo)
		pytest.raises(NotImplementedError, WCacheStorage.clear, None, None)
		pytest.raises(NotImplementedError, WCacheStorage.get_cache, None, foo)

		class Storage(WCacheStorage):

			__cache_value__ = WCacheStorage.CacheEntry(has_value=True, cached_value='zzz')

			def put(self, result, decorated_function, *args, **kwargs):
				pass

			def clear(self, decorated_function=None):
				pass

			def get_cache(self, decorated_function, *args, **kwargs):
				return self.__cache_value__

		storage = Storage()
		assert(storage.has(foo) is True)
		assert(storage.get_result(foo) == 'zzz')

		storage.__cache_value__ = WCacheStorage.CacheEntry()
		assert(storage.has(foo) is False)
		pytest.raises(WCacheStorage.CacheMissedException, storage.get_result, foo)


class TestWGlobalSingletonCacheStorage:

	def test(self):
		foo = TestWCacheStorage.foo
		bar = TestWCacheStorage.bar

		global_singleton_storage = WGlobalSingletonCacheStorage()
		pytest.raises(WCacheStorage.CacheMissedException, global_singleton_storage.get_result, foo)
		assert(global_singleton_storage.has(foo) is False)
		assert(global_singleton_storage.has(foo, 1, k=2) is False)
		assert(global_singleton_storage.has(bar) is False)
		assert(global_singleton_storage.has(bar, 2, m='s') is False)

		global_singleton_storage.put(4, foo)
		assert(global_singleton_storage.has(foo) is True)
		assert(global_singleton_storage.get_result(foo) == 4)
		assert(global_singleton_storage.has(foo, 1, k=2) is True)
		assert(global_singleton_storage.get_result(foo, 1, k=2) == 4)
		assert(global_singleton_storage.has(bar) is False)
		assert(global_singleton_storage.has(bar, 2, m='s') is False)

		global_singleton_storage.put('q', bar, 7, l=8)
		assert(global_singleton_storage.has(bar) is True)
		assert(global_singleton_storage.get_result(bar) == 'q')
		assert(global_singleton_storage.has(bar, 2, m='s') is True)
		assert(global_singleton_storage.get_result(bar, 2, m='s') == 'q')

		global_singleton_storage.put('z', bar)
		assert(global_singleton_storage.get_result(bar) == 'z')
		assert(global_singleton_storage.get_result(bar, 2, m='s') == 'z')

		global_singleton_storage.clear(bar)
		assert(global_singleton_storage.has(foo) is True)
		assert(global_singleton_storage.get_result(foo) == 4)
		assert(global_singleton_storage.has(foo, 1, k=2) is True)
		assert(global_singleton_storage.get_result(foo, 1, k=2) == 4)
		assert(global_singleton_storage.has(bar) is False)
		assert(global_singleton_storage.has(bar, 2, m='s') is False)

		global_singleton_storage.clear()
		assert(global_singleton_storage.has(foo) is False)
		assert(global_singleton_storage.has(foo, 1, k=2) is False)
		assert(global_singleton_storage.has(bar) is False)
		assert(global_singleton_storage.has(bar, 2, m='s') is False)


class TestWInstanceSingletonCacheStorage:

	def test_cache_record(self):
		foo = TestWCacheStorage.foo

		cache_record = WInstanceSingletonCacheStorage.InstanceCacheRecord('qaz', foo)
		assert(cache_record.decorated_function() == foo)
		cache_entry = cache_record.cache_entry()
		assert(isinstance(cache_entry, WCacheStorage.CacheEntry) is True)
		assert(cache_entry.has_value is True)
		assert(cache_entry.cached_value == 'qaz')

		cache_record.update('wsx')
		assert(cache_record.cache_entry().cached_value == 'wsx')

		cache_record = WInstanceSingletonCacheStorage.InstanceCacheRecord.create('zxc', foo)
		assert(isinstance(cache_record, WInstanceSingletonCacheStorage.InstanceCacheRecord) is True)
		assert(cache_record.cache_entry().has_value is True)
		assert(cache_record.cache_entry().cached_value == 'zxc')

	def test(self):
		foo = TestWCacheStorage.foo

		instance_singleton_storage = WInstanceSingletonCacheStorage()
		pytest.raises(RuntimeError, instance_singleton_storage.put, 1, foo)
		pytest.raises(RuntimeError, instance_singleton_storage.has, foo)
		pytest.raises(RuntimeError, instance_singleton_storage.get_result, foo)

		class A:
			def foo(self):
				pass

			def bar(self):
				pass

		a1 = A()
		a2 = A()
		a3 = A()

		pytest.raises(WCacheStorage.CacheMissedException, instance_singleton_storage.get_result, A.foo, a1)
		assert(instance_singleton_storage.has(A.foo, a1) is False)
		assert(instance_singleton_storage.has(A.foo, a2) is False)
		assert(instance_singleton_storage.has(A.foo, a1, 1, k=2) is False)
		assert(instance_singleton_storage.has(A.foo, a2, 1, k=2) is False)
		assert(instance_singleton_storage.has(A.bar, a1) is False)
		assert(instance_singleton_storage.has(A.bar, a2) is False)
		assert(instance_singleton_storage.has(A.bar, a1, m='s') is False)
		assert(instance_singleton_storage.has(A.bar, a2, m='s') is False)

		instance_singleton_storage.put(8, A.foo, a1)
		assert(instance_singleton_storage.has(A.foo, a1) is True)
		assert(instance_singleton_storage.get_result(A.foo, a1) == 8)
		assert(instance_singleton_storage.has(A.foo, a2) is False)
		assert(instance_singleton_storage.has(A.foo, a1, 1, k=2) is True)
		assert(instance_singleton_storage.get_result(A.foo, a1, 1, k=2) == 8)
		assert(instance_singleton_storage.has(A.foo, a2, 1, k=2) is False)
		assert(instance_singleton_storage.has(A.bar, a1) is False)
		assert(instance_singleton_storage.has(A.bar, a2) is False)
		assert(instance_singleton_storage.has(A.bar, a1, m='s') is False)
		assert(instance_singleton_storage.has(A.bar, a2, m='s') is False)

		instance_singleton_storage.put(18, A.bar, a2, k=9)
		assert(instance_singleton_storage.has(A.foo, a1) is True)
		assert(instance_singleton_storage.get_result(A.foo, a1) == 8)
		assert(instance_singleton_storage.has(A.foo, a2) is False)
		assert(instance_singleton_storage.has(A.foo, a1, 1, k=2) is True)
		assert(instance_singleton_storage.get_result(A.foo, a1, 1, k=2) == 8)
		assert(instance_singleton_storage.has(A.foo, a2, 1, k=2) is False)
		assert(instance_singleton_storage.has(A.bar, a1) is False)
		assert(instance_singleton_storage.has(A.bar, a2) is True)
		assert(instance_singleton_storage.get_result(A.bar, a2) == 18)
		assert(instance_singleton_storage.has(A.bar, a1, m='s') is False)
		assert(instance_singleton_storage.has(A.bar, a2, m='s') is True)
		assert(instance_singleton_storage.get_result(A.bar, a2, m='s') == 18)

		instance_singleton_storage.put(7, A.bar, a1, q=5)
		assert(instance_singleton_storage.has(A.foo, a1) is True)
		assert(instance_singleton_storage.get_result(A.foo, a1) == 8)
		assert(instance_singleton_storage.has(A.foo, a2) is False)
		assert(instance_singleton_storage.has(A.foo, a1, 1, k=2) is True)
		assert(instance_singleton_storage.get_result(A.foo, a1, 1, k=2) == 8)
		assert(instance_singleton_storage.has(A.foo, a2, 1, k=2) is False)
		assert(instance_singleton_storage.has(A.bar, a1) is True)
		assert(instance_singleton_storage.get_result(A.bar, a1) == 7)
		assert(instance_singleton_storage.has(A.bar, a2) is True)
		assert(instance_singleton_storage.get_result(A.bar, a2) == 18)
		assert(instance_singleton_storage.has(A.bar, a1, m='s') is True)
		assert(instance_singleton_storage.get_result(A.bar, a1, m='s') == 7)
		assert(instance_singleton_storage.has(A.bar, a2, m='s') is True)
		assert(instance_singleton_storage.get_result(A.bar, a2, m='s') == 18)

		instance_singleton_storage.put('d', A.bar, a2)
		assert(instance_singleton_storage.get_result(A.bar, a2) == 'd')
		assert(instance_singleton_storage.get_result(A.bar, a2, m='s') == 'd')

		instance_singleton_storage.clear(decorated_function=A.bar)
		assert(instance_singleton_storage.has(A.foo, a1) is True)
		assert(instance_singleton_storage.has(A.foo, a2) is False)
		assert(instance_singleton_storage.has(A.bar, a1) is False)
		assert(instance_singleton_storage.has(A.bar, a2) is False)

		instance_singleton_storage.clear()
		assert(instance_singleton_storage.has(A.foo, a1) is False)
		assert(instance_singleton_storage.has(A.foo, a2) is False)
		assert(instance_singleton_storage.has(A.bar, a1) is False)
		assert(instance_singleton_storage.has(A.bar, a2) is False)

		pytest.raises(WCacheStorage.CacheMissedException, instance_singleton_storage.get_result, A.foo, a3)

		class DummyCacheRecord(WInstanceSingletonCacheStorage.InstanceCacheRecord):

			def cache_entry(self, *args, **kwargs):
				return WCacheStorage.CacheEntry()

		pytest.raises(TypeError, WInstanceSingletonCacheStorage, cache_record_cls=A)
		instance_singleton_storage = WInstanceSingletonCacheStorage(cache_record_cls=DummyCacheRecord)

		instance_singleton_storage.put('d', A.bar, a2)
		assert(instance_singleton_storage.has(A.bar, a2) is False)

	def test_get_cache(self):

		class A:
			def foo(self):
				pass

		a1 = A()
		a2 = A()
		a3 = A()

		instance_singleton_storage = WInstanceSingletonCacheStorage()
		assert(instance_singleton_storage.cache_hit() is None)
		assert(instance_singleton_storage.cache_missed() is None)

		assert(instance_singleton_storage.get_cache(A.foo, a1).has_value is False)
		assert(instance_singleton_storage.cache_hit() is None)
		assert(instance_singleton_storage.cache_missed() is None)

		instance_singleton_storage.put(1, A.foo, a1)
		assert(instance_singleton_storage.get_cache(A.foo, a1).has_value is True)
		assert(instance_singleton_storage.cache_hit() is None)
		assert(instance_singleton_storage.cache_missed() is None)

		instance_singleton_storage = WInstanceSingletonCacheStorage(statistic=True)
		assert(instance_singleton_storage.cache_hit() == 0)
		assert(instance_singleton_storage.cache_missed() == 0)

		assert(instance_singleton_storage.get_cache(A.foo, a1).has_value is False)
		assert(instance_singleton_storage.cache_hit() == 0)
		assert(instance_singleton_storage.cache_missed() == 1)

		assert(instance_singleton_storage.get_cache(A.foo, a2).has_value is False)
		assert(instance_singleton_storage.cache_hit() == 0)
		assert(instance_singleton_storage.cache_missed() == 2)

		instance_singleton_storage.put(1, A.foo, a1)
		assert(instance_singleton_storage.get_cache(A.foo, a1).has_value is True)
		assert(instance_singleton_storage.cache_hit() == 1)
		assert(instance_singleton_storage.cache_missed() == 2)

		assert(instance_singleton_storage.get_cache(A.foo, a3).has_value is False)
		assert(instance_singleton_storage.cache_hit() == 1)
		assert(instance_singleton_storage.cache_missed() == 3)

		instance_singleton_storage.put(1, A.foo, a3)
		assert(instance_singleton_storage.get_cache(A.foo, a3).has_value is True)
		assert(instance_singleton_storage.cache_hit() == 2)
		assert(instance_singleton_storage.cache_missed() == 3)

		assert(instance_singleton_storage.get_cache(A.foo, a3).has_value is True)
		assert(instance_singleton_storage.cache_hit() == 3)
		assert(instance_singleton_storage.cache_missed() == 3)

		assert(instance_singleton_storage.get_cache(A.foo, a3).has_value is True)
		assert(instance_singleton_storage.cache_hit() == 4)
		assert(instance_singleton_storage.cache_missed() == 3)

		instance_singleton_storage.clear()
		assert(instance_singleton_storage.cache_hit() == 0)
		assert(instance_singleton_storage.cache_missed() == 0)

		assert(instance_singleton_storage.get_cache(A.foo, a3).has_value is False)
		assert(instance_singleton_storage.cache_hit() == 0)
		assert(instance_singleton_storage.cache_missed() == 1)

		class DummyCacheRecord(WInstanceSingletonCacheStorage.InstanceCacheRecord):

			def cache_entry(self, *args, **kwargs):
				return WCacheStorage.CacheEntry()

		instance_singleton_storage = WInstanceSingletonCacheStorage(
			cache_record_cls=DummyCacheRecord, statistic=True
		)
		assert(instance_singleton_storage.cache_hit() == 0)
		assert(instance_singleton_storage.cache_missed() == 0)

		instance_singleton_storage.put(1, A.foo, a1)
		assert(instance_singleton_storage.get_cache(A.foo, a1).has_value is False)
		assert(instance_singleton_storage.cache_hit() == 0)
		assert(instance_singleton_storage.cache_missed() == 1)

		assert(instance_singleton_storage.get_cache(A.foo, a2).has_value is False)
		assert(instance_singleton_storage.cache_hit() == 0)
		assert(instance_singleton_storage.cache_missed() == 2)


def test_cache_control():

	def foo(a, b=None):
		return a + (0 if b is None else b)

	validator_trigger = True

	def validator(fn, *args, **kwargs):
		return validator_trigger

	decorated_foo = cache_control()(foo)
	assert(decorated_foo(1, b=2) == 3)
	assert(decorated_foo(1, b=5) == 3)
	assert(decorated_foo(7) == 3)

	decorated_foo = cache_control(validator=validator)(foo)
	assert(decorated_foo(8) == 8)
	assert(decorated_foo(1, b=2) == 8)
	assert(decorated_foo(1, b=5) == 8)
	assert(decorated_foo(7) == 8)

	validator_trigger = False
	assert(decorated_foo(1, b=2) == 3)
	assert(decorated_foo(1, b=5) == 6)
	assert(decorated_foo(7) == 7)

	validator_trigger = True
	assert(decorated_foo(1, b=2) == 7)
	assert(decorated_foo(1, b=5) == 7)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.os.linux.lvm import WLVMInfoCommand, WLVMInfo, WPhysicalVolume, WVolumeGroup, WLogicalVolume


class TestWLVMInfoCommand:

	def test(self):
		info_command = WLVMInfoCommand('command1', 10)
		assert(isinstance(info_command, WLVMInfoCommand) is True)
		assert(info_command.command() == 'command1')
		assert(info_command.fields_count() == 10)
		assert(info_command.cmd_timeout() > 0)
		assert(info_command.sudo() is False)

		info_command = WLVMInfoCommand('command2', 7, cmd_timeout=1, sudo=True)
		assert(isinstance(info_command, WLVMInfoCommand) is True)
		assert(info_command.command() == 'command2')
		assert(info_command.fields_count() == 7)
		assert(info_command.cmd_timeout() == 1)
		assert(info_command.sudo() is True)


class TestWLVMInfo:

	def test(self):
		lvm_info = WLVMInfo('command1', 3, ['field1', 'option2', 'value3'])
		assert(isinstance(lvm_info, WLVMInfo) is True)
		assert(lvm_info.lvm_entity() == ('field1', 'option2', 'value3'))
		assert(isinstance(lvm_info.lvm_command(), WLVMInfoCommand) is True)
		assert(lvm_info.lvm_command().command() == 'command1')
		assert (lvm_info.lvm_command().fields_count() == 3)
		assert (lvm_info.lvm_command().cmd_timeout() == WLVMInfo.__lvm_info_cmd_timeout__)
		assert (lvm_info.lvm_command().sudo() is False)

		class TestLVMInfo(WLVMInfo):
			__lvm_info_cmd_timeout__ = 10

		lvm_info = TestLVMInfo('command2', 4, ['field1', 'option2', 'value3', ''], sudo=True)
		assert(lvm_info.lvm_command().command() == 'command2')
		assert (lvm_info.lvm_command().fields_count() == 4)
		assert (lvm_info.lvm_command().cmd_timeout() == TestLVMInfo.__lvm_info_cmd_timeout__)
		assert (lvm_info.lvm_command().sudo() is True)

		pytest.raises(ValueError, WLVMInfo, 'command', 2, [])


class TestWPhysicalVolume:
	def test(self):
		example = \
			'/dev/sda1:test-vg:16287744:-1:8:8:-1:4096:1988:200:1788:GHskdS-776l-mFbq-UDkG-MqN9-rYTO-rNpHGy'
		phy_vol = WPhysicalVolume(example.split(':'))
		assert(isinstance(phy_vol, WPhysicalVolume) is True)
		assert(isinstance(phy_vol, WLVMInfo) is True)
		assert(phy_vol.lvm_command().sudo() is False)
		assert(phy_vol.device_name() == '/dev/sda1')
		assert(phy_vol.volume_group() == 'test-vg')
		assert(phy_vol.sectors_count() == 16287744)
		assert(phy_vol.extent_size() == 4096)
		assert(phy_vol.total_extents() == 1988)
		assert(phy_vol.free_extents() == 200)
		assert(phy_vol.allocated_extents() == 1788)
		assert(phy_vol.uuid() == 'GHskdS-776l-mFbq-UDkG-MqN9-rYTO-rNpHGy')

		phy_vol = WPhysicalVolume(example.split(':'), sudo=True)
		assert(phy_vol.lvm_command().sudo() is True)


class TestWVolumeGroup:

	def test(self):
		example = 'test-vg:r/w:772:-1:0:2:1:-1:0:1:1:8142848:4096:1988:1788:200:' \
			'KupnMx-Fizr-DxTk-2GTf-9ELZ-J4u9-82Muq7'
		vol_group = WVolumeGroup(example.split(':'))
		assert(isinstance(vol_group, WVolumeGroup) is True)
		assert(isinstance(vol_group, WLVMInfo) is True)
		assert(vol_group.lvm_command().sudo() is False)

		assert(vol_group.group_name() == 'test-vg')
		assert(vol_group.group_access() == 'r/w')
		assert(vol_group.maximum_logical_volumes() == 0)
		assert(vol_group.logical_volumes() == 2)
		assert(vol_group.opened_logical_volumes() == 1)
		assert(vol_group.maximum_physical_volumes() == 0)
		assert(vol_group.physical_volumes() == 1)
		assert(vol_group.maximum_physical_volumes() == 0)
		assert(vol_group.actual_volumes() == 1)
		assert(vol_group.size() == 8142848)
		assert(vol_group.extent_size() == 4096)
		assert(vol_group.total_extents() == 1988)
		assert(vol_group.allocated_extents() == 1788)
		assert(vol_group.free_extents() == 200)
		assert(vol_group.uuid() == 'KupnMx-Fizr-DxTk-2GTf-9ELZ-J4u9-82Muq7')

		vol_group = WVolumeGroup(example.split(':'), sudo=True)
		assert (vol_group.lvm_command().sudo() is True)


class TestWLogicalVolume:

	def test(self):
		example = '/dev/test-vg/test-lv:test-vg:3:1:-1:1:14647296:1788:-1:0:-1:254:0'
		log_vol = WLogicalVolume(example.split(':'))
		assert(isinstance(log_vol, WLogicalVolume) is True)
		assert(isinstance(log_vol, WLVMInfo) is True)
		assert(log_vol.lvm_command().sudo() is False)

		assert(log_vol.volume_path() == '/dev/test-vg/test-lv')
		assert(log_vol.volume_name() == 'test-lv')
		assert(log_vol.volume_group_name() == 'test-vg')
		assert(log_vol.sectors_count() == 14647296)
		assert(log_vol.extents_count() == 1788)
		assert (log_vol.device_number() == (254, 0))

		log_vol = WLogicalVolume(example.split(':'), sudo=True)
		assert(log_vol.lvm_command().sudo() is True)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.primitives import WIPV4SocketInfo
from wasp_general.network.web.headers import WHTTPHeaders
from wasp_general.network.web.proto import WWebSessionProto
from wasp_general.network.web.request import WWebRequest


class TestWWebRequest:

	class Session(WWebSessionProto):

		def client_address(self):
			WIPV4SocketInfo('127.0.0.1', 65000)

		def server_address(self):
			WIPV4SocketInfo('127.0.0.1', 80)

		def protocol(self):
			return 'http'

		def protocol_version(self):
			return '0.9'

		def read_request(self):
			pass

		def write_response(self, request, response):
			pass

		def session_close(self):
			pass

	def test_request(self):
		session = TestWWebRequest.Session()
		request = WWebRequest(session, 'GET', '/foo/bar')
		assert(isinstance(request, WWebRequest) is True)
		assert(request.session() == session)
		assert(request.method() == 'GET')
		assert(request.path() == '/foo/bar')
		assert(request.headers() is None)
		assert(request.request_data() is None)

		headers1 = WHTTPHeaders(header1='value1')
		request = WWebRequest(session, 'GET', '/foo/bar', headers=headers1, request_data=b'request payload')
		assert(request.headers()['header1'] == ('value1',))
		assert(request.request_data() == b'request payload')

		request.set_headers(headers=WHTTPHeaders(header1='value2'))
		assert(request.headers()['header1'] == ('value2',))
		request.set_request_data(b'payload')
		assert(request.request_data() == b'payload')

		request = WWebRequest.parse_request_line(session, 'GET /zzz')
		assert(request.method() == 'GET')
		assert(request.path() == '/zzz')

		request = WWebRequest.parse_request_line(session, 'POST /foo HTTP/1.0')
		assert(request.method() == 'POST')
		assert(request.path() == '/foo')
		pytest.raises(ValueError, WWebRequest.parse_request_line, session, 'zzz')

		request.parse_headers('Header-X: foo\r\nHeader-Y: bar\r\n')
		assert(request.headers()['Header-X'] == ('foo',))
		assert(request.headers()['Header-Y'] == ('bar',))

		ro_request = request.ro()
		assert(ro_request.method() == 'POST')
		assert(ro_request.path() == '/foo')
		assert(ro_request.headers()['Header-X'] == ('foo',))
		assert(ro_request.headers()['Header-Y'] == ('bar',))
		pytest.raises(RuntimeError, ro_request.set_headers, WHTTPHeaders())
		pytest.raises(RuntimeError, ro_request.set_request_data, b'')
		pytest.raises(RuntimeError, ro_request.parse_headers, 'Headers-X: foo\r\n')



# -*- coding: utf-8 -*-
# wasp_general/task/dependency.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type, verify_subclass
from wasp_general.task.registry import WTaskRegistry, WRegisteredTask, WTaskRegistryStorage
from wasp_general.task.base import WTask, WStoppableTask


class WDependentTask(WRegisteredTask):
	""" Metaclass for dependent tasks. It is used for automatic resolving required dependencies (starting required
	tasks). Derived classes must be able to be constructed with constructor without arguments or they must
	override :meth:`.WDependentTask.start_dependent_task` method.

	If derived class inherits :class:`wasp_general.task.base.WStoppableTask` class, then it could be stopped
	(automatically stopped via registry class, such as :class:`.WTaskDependencyRegistry`)

	When __registry_tag__ is set to None, related task registry must have '__skip_none_registry_tag__' flag
	set to False, otherwise - exception is raised. Such task (that have __registry_tag__ set to None) will be
	excluded from registry.

	In order to include task to registry, task must define __registry_tag__, which has to be a str class.
	"""

	__dependency__ = []
	""" List of tags (str). Each tag represent task, that required to start prior to this task.
	"""
	__description__ = None
	""" Just task description. Should be str type.
	"""

	def __init__(cls, name, bases, namespace):
		""" Construct new class. Derived class must redefine __registry__ and __registry_tag__ properties.
		In order to start dependency task automatically, property __dependency__ must be redefined.

		It is highly important, that derived class method :meth:`.WDependentTask.start_dependent_task` will
		be able to construct and start this task.

		:param name: as name in type(cls, name, bases, namespace)
		:param bases: as bases in type(cls, name, bases, namespace)
		:param namespace: as namespace in type(cls, name, bases, namespace)
		"""

		if cls.__auto_registry__ is True:

			if cls.__registry__ is None:
				raise ValueError('__registry__ must be defined')

			if issubclass(cls.__registry__, WTaskDependencyRegistry) is False:
				raise TypeError(
					"Property '__registry__' of tasks class has invalid type (must be "
					"WTaskDependencyRegistry or its subclass)"
				)

			if cls.__registry_tag__ is None:
				if cls.__registry__.__skip_none_registry_tag__ is False:
					raise ValueError("Property '__registry_tag__' must be defined")
			elif isinstance(cls.__registry_tag__, str) is False:
				raise TypeError("Property '__registry_tag__' must be string type")

		WRegisteredTask.__init__(cls, name, bases, namespace)

	def start_dependent_task(cls):
		""" Start this task and return its instance

		:return: WTask
		"""
		task = cls()
		# noinspection PyUnresolvedReferences
		task.start()
		return task


class WTaskDependencyRegistryStorage(WTaskRegistryStorage):
	""" Storage that is used to store :class:`.WDependentTask` task.
	"""

	__multiple_tasks_per_tag__ = False
	""" Each task must have unique __registry_tag__ property.
	See :attr:`wasp_general.task.registry.WTaskRegistryStorage.__multiple_tasks_per_tag__`
	"""

	def __init__(self):
		""" Construct new storage
		"""
		WTaskRegistryStorage.__init__(self)
		self.__started = []

	@verify_type('paranoid', task_cls=WDependentTask)
	def add(self, task_cls):
		""" Add task to this storage. Multiple tasks with the same tag are not allowed

		:param task_cls: task to add
		:return: None
		"""
		return WTaskRegistryStorage.add(self, task_cls)

	@verify_subclass(task_cls=WTask)
	@verify_type(task_cls=WDependentTask)
	def dependency_check(self, task_cls, skip_unresolved=False):
		""" Check dependency of task for irresolvable conflicts (like task to task mutual dependency)

		:param task_cls: task to check
		:param skip_unresolved: flag controls this method behaviour for tasks that could not be found. \
		When False, method will raise an exception if task tag was set in dependency and the related task \
		wasn't found in registry. When True that unresolvable task will be omitted

		:return: None
		"""

		def check(check_task_cls, global_dependencies):
			if check_task_cls.__registry_tag__ in global_dependencies:
				raise RuntimeError('Recursion dependencies for %s' % task_cls.__registry_tag__)

			dependencies = global_dependencies.copy()
			dependencies.append(check_task_cls.__registry_tag__)

			for dependency in check_task_cls.__dependency__:
				dependent_task = self.tasks_by_tag(dependency)
				if dependent_task is None and skip_unresolved is False:
					raise RuntimeError(
						"Task '%s' dependency unresolved (%s)" %
						(task_cls.__registry_tag__, dependency)
					)

				if dependent_task is not None:
					check(dependent_task, dependencies)

		check(task_cls, [])

	def started_tasks(self, task_registry_id=None, task_cls=None):
		""" Return tasks that was started. Result way be filtered by the given arguments.

		:param task_registry_id: if it is specified, then try to return single task which id is the same as \
		this value.
		:param task_cls: if it is specified then result will be consists of this subclass only

		:return: None or WTask or tuple of WTask
		"""

		if task_registry_id is not None:
			task = None
			for registered_task in self.__started:
				if registered_task.__registry_tag__ == task_registry_id:
					task = registered_task

			if task_cls is not None and task is not None:
				if isinstance(task, task_cls) is True:
					return task
				return None
			return task

		result = filter(lambda x: x is not None, self.__started)
		if task_cls is not None:
			result = filter(lambda x: isinstance(x, task_cls), result)
		return tuple(result)

	@verify_type(task_tag=str, skip_unresolved=bool)
	def start_task(self, task_tag, skip_unresolved=False):
		""" Check dependency for the given task_tag and start task. For dependency checking see
		:meth:`.WTaskDependencyRegistryStorage.dependency_check`. If task is already started then it must be
		stopped before it will be started again.

		:param task_tag: task to start. Any required dependencies will be started automatically.
		:param skip_unresolved: flag controls this method behaviour for tasks that could not be found. \
		When False, method will raise an exception if task tag was set in dependency and the related task \
		wasn't found in registry. When True that unresolvable task will be omitted

		:return: None
		"""
		if self.started_tasks(task_registry_id=task_tag) is not None:
			return

		task_cls = self.tasks_by_tag(task_tag)
		if task_cls is None:
			raise RuntimeError("Task '%s' wasn't found" % task_tag)

		self.dependency_check(task_cls, skip_unresolved=skip_unresolved)

		def start_dependency(start_task_cls):
			for dependency in start_task_cls.__dependency__:

				if self.started_tasks(task_registry_id=dependency) is not None:
					continue

				dependent_task = self.tasks_by_tag(dependency)

				if dependent_task is not None:
					start_dependency(dependent_task)

			self.__started.append(start_task_cls.start_dependent_task())

		start_dependency(task_cls)

	@verify_type(task_tag=str, stop_dependent=bool, stop_requirements=bool)
	def stop_task(self, task_tag, stop_dependent=True, stop_requirements=False):
		""" Stop task with the given task tag. If task already stopped, then nothing happens.

		:param task_tag: task to stop
		:param stop_dependent: if True, then every task, that require the given task as dependency, will be \
		stopped before.
		:param stop_requirements: if True, then every task, that is required as dependency for the given task, \
		will be stopped after.
		:return: None
		"""
		# TODO: "coverage" requires more tests

		task = self.started_tasks(task_registry_id=task_tag)

		if task is None:
			return

		def stop(task_to_stop):
			if task_to_stop in self.__started:
				if isinstance(task_to_stop, WStoppableTask) is True:
					task_to_stop.stop()
				self.__started.remove(task_to_stop)

		def stop_dependency(task_to_stop):
			deeper_dependencies = []
			for dependent_task in self.__started:
				if task_to_stop.__registry_tag__ in dependent_task.__class__.__dependency__:
					deeper_dependencies.append(dependent_task)

			for dependent_task in deeper_dependencies:
				stop_dependency(dependent_task)

			stop(task_to_stop)

		def calculate_requirements(task_to_stop, cross_requirements=False):
			requirements = set()

			for dependent_task in self.__started:
				if dependent_task.__class__.__registry_tag__ in task_to_stop.__class__.__dependency__:
					requirements.add(dependent_task)

			if cross_requirements is True:
				return requirements

			result = set()
			for task_a in requirements:
				requirement_match = False
				for task_b in requirements:
					if task_a.__class__.__registry_tag__ in task_b.__class__.__dependency__:
						requirement_match = True
						break
				if requirement_match is False:
					result.add(task_a)
			return result

		def calculate_priorities(task_to_stop, *extra_tasks, current_result=None, requirements_left=None):
			if current_result is None:
				current_result = []

			tasks_to_stop = [task_to_stop]
			if len(extra_tasks) > 0:
				tasks_to_stop.extend(extra_tasks)

			current_result.append(list(tasks_to_stop))

			all_requirements = calculate_requirements(tasks_to_stop[0], cross_requirements=True)
			nested_requirements = calculate_requirements(tasks_to_stop[0])

			for dependent_task in tasks_to_stop[1:]:
				nested_requirements = nested_requirements.union(calculate_requirements(dependent_task))
				all_requirements.update(calculate_requirements(dependent_task, cross_requirements=True))

			all_requirements = all_requirements.difference(nested_requirements)

			if requirements_left is not None:
				requirements_left = requirements_left.difference(all_requirements)
				nested_requirements.update(requirements_left)

			if len(nested_requirements) == 0:
				return current_result

			return calculate_priorities(
				*list(nested_requirements), current_result=current_result, requirements_left=all_requirements
			)

		if stop_dependent is True:
			stop_dependency(task)

		if stop_requirements is True:
			for task_list in calculate_priorities(task):
				for single_task in task_list:
					stop(single_task)

		if stop_dependent is not True:  # check if we've already stopped this task
			stop(task)


class WTaskDependencyRegistry(WTaskRegistry):
	""" Registry for the :class:`.WDependentTask` classes. Registry storage must be

	Derived classes must redefine __registry_storage__ property (which has to be
	:class:`.WTaskDependencyRegistryStorage` instance). (see :attr:`.WTaskRegistry.__registry_storage__`)
	"""

	@classmethod
	def registry_storage(cls):
		""" Get registry storage

		:return: WTaskDependencyRegistryStorage
		"""
		if cls.__registry_storage__ is None:
			raise ValueError('__registry_storage__ must be defined')
		if isinstance(cls.__registry_storage__, WTaskDependencyRegistryStorage) is False:
			raise TypeError(
				"Property '__registry_storage__' is invalid (must derived from WTaskRegistryBase)"
			)

		return cls.__registry_storage__

	@classmethod
	def start_task(cls, task_tag, skip_unresolved=False):
		""" Start task from registry

		:param task_tag: same as in :meth:`.WTaskDependencyRegistryStorage.start_task` method
		:param skip_unresolved: same as in :meth:`.WTaskDependencyRegistryStorage.start_task` method
		:return: None
		"""
		registry = cls.registry_storage()
		registry.start_task(task_tag, skip_unresolved=skip_unresolved)

	@classmethod
	def stop_task(cls, task_tag, stop_dependent=True, stop_requirements=False):
		""" Stop started task from registry

		:param task_tag: same as in :meth:`.WTaskDependencyRegistryStorage.stop_task` method
		:param stop_dependent: same as in :meth:`.WTaskDependencyRegistryStorage.stop_task` method
		:param stop_requirements: same as in :meth:`.WTaskDependencyRegistryStorage.stop_task` method
		:return: None
		"""
		registry = cls.registry_storage()
		registry.stop_task(task_tag, stop_dependent=stop_dependent, stop_requirements=stop_requirements)

	@classmethod
	def all_stop(cls):
		""" Stop every task started within this registry

		:return: None
		"""
		for task_tag in cls.registry_storage().tags():
			cls.stop_task(task_tag)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.crypto.hex import WHex, WUnHex


class TestWHex:
	def test___init__(self):
		pytest.raises(TypeError, WHex)
		pytest.raises(TypeError, WHex, 1)
		pytest.raises(TypeError, WHex, 'asd')
		assert(isinstance(WHex(b'asd'), WHex) is True)

		pytest.raises(TypeError, WUnHex)
		pytest.raises(TypeError, WUnHex, 1)
		assert(isinstance(WUnHex('asd'), WUnHex) is True)
		assert(isinstance(WUnHex(b'asd'), WUnHex) is True)

		assert(isinstance(str(WHex(b'a')), str) is True)
		assert(isinstance(bytes(WUnHex('61')), bytes) is True)
		assert(isinstance(bytes(WUnHex(b'61')), bytes) is True)

		assert(str(WHex(b'a')) == '61')
		assert(bytes(WUnHex('61')) == b'a')
		assert(bytes(WUnHex(b'61')) == b'a')

		assert(bytes(WUnHex(str(WHex(b'qwerty')))) == b'qwerty')



# -*- coding: utf-8 -*-
# wasp_general/<FILENAME>.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod, abstractclassmethod
from Crypto.Hash import SHA as SHA1, SHA224, SHA256, SHA384, SHA512, MD5

from wasp_general.verify import verify_type


class WHashGeneratorProto(metaclass=ABCMeta):
	""" Prototype for hash-generator.

	note: there is commonly used feature that most hash generator objects have - digest_size attribute. So it is
	better to create this attribute.
	"""

	@abstractmethod
	@verify_type(data=bytes)
	def update(self, data):
		""" Update digest by hashing the specified data

		:param data: data to hash

		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def digest(self):
		""" Return current digest

		:return: bytes
		"""
		raise NotImplementedError('This method is abstract')

	def hexdigest(self):
		""" Return current digest in hex-alike string

		:return: str
		"""
		return ''.join(["{:02x}".format(x).upper() for x in self.digest()])

	@abstractclassmethod
	def generator_digest_size(cls):
		""" Return generator digest size

		:return: int
		"""
		raise NotImplementedError('This method is abstract')

	@abstractclassmethod
	def generator_name(cls):
		""" Return hash-function name

		:return: str
		"""
		raise NotImplementedError('This method is abstract')

	@abstractclassmethod
	def generator_family(cls):
		""" Return name of hash-function family (like: 'SHA')

		:return: str or None (if no available)
		"""
		raise NotImplementedError('This method is abstract')

	@abstractclassmethod
	@verify_type(data=(bytes, None))
	def new(cls, data=None):
		""" Return new generator and hash the specified data (if defined)

		:param data: data to hash

		:return: WHashGeneratorProto
		"""
		raise NotImplementedError('This method is abstract')


class WPyCryptoHashAdapter(WHashGeneratorProto):
	""" Class that adapts the specified PyCrypto hashing class to WHashGeneratorProto implementation
	"""

	__pycrypto_cls__ = None
	""" PyCrypto class to adapt. Must be override in a derived classes
	"""

	__generator_name__ = None
	""" Hash-function name (like 'SHA1' or 'MD5'). Must be override in a derived classes
	"""

	__generator_family__ = None
	""" Hash-function family name (like: 'SHA')
	"""

	def __init__(self):
		""" Create new hash generator
		"""
		WHashGeneratorProto.__init__(self)

		if self.__class__.__pycrypto_cls__ is None:
			raise ValueError('"__pycrypto_cls__" must be override in a derived class')

		self.__pycrypto_obj = self.__class__.__pycrypto_cls__.new()
		self.digest_size = self.__class__.generator_digest_size()

	@verify_type(data=bytes)
	def update(self, data):
		""" :meth:`.WHashGeneratorProto.update` implementation
		"""
		self.__pycrypto_obj.update(data)

	def digest(self):
		""" :meth:`.WHashGeneratorProto.digest` implementation
		"""
		return self.__pycrypto_obj.digest()

	def pycrypto(self):
		""" In rare cases original PyCrypto object is required. In most cases this method should be avoided,
		as it is can be removed at any time.

		One of an example of this method usage is PyCrypto HMAC (and so :class:`.WHMAC`). They require, that
		hash-generator object must have "copy" method to be implemented. But I have not found a way to make
		HMAC work.

		:return: PyCrypto Hash object
		"""
		return self.__pycrypto_obj

	@classmethod
	def generator_digest_size(cls):
		""" :meth:`.WHashGeneratorProto.generator_digest_size` implementation
		"""
		if cls.__pycrypto_cls__ is None:
			raise ValueError('"__pycrypto_cls__" must be override in a derived class')
		return cls.__pycrypto_cls__.digest_size

	@classmethod
	def generator_name(cls):
		""" :meth:`.WHashGeneratorProto.generator_name` implementation
		"""
		if cls.__generator_name__ is None:
			raise ValueError('"__generator_name__" should be override in a derived class')
		if isinstance(cls.__generator_name__, str) is False:
			raise TypeError('"__generator_name__" should be a str instance')
		return cls.__generator_name__.upper()

	@classmethod
	def generator_family(cls):
		""" :meth:`.WHashGeneratorProto.generator_family` implementation
		"""
		if cls.__generator_family__ is not None:
			if isinstance(cls.__generator_family__, str) is False:
				raise TypeError('"__generator_class__"  if defined must be a str instance')

		if cls.__generator_family__ is not None:
			return cls.__generator_family__.upper()

	@classmethod
	@verify_type(data=(bytes, None))
	def new(cls, data=None):
		""" :meth:`.WHashGeneratorProto.new` implementation
		"""
		obj = cls()
		if data is not None:
			obj.update(data)
		return obj


class WSHAFamily(WPyCryptoHashAdapter):
	""" Class that represent SHA-family hash-generators
	"""
	__generator_family__ = 'SHA'


class WSHA1(WSHAFamily):
	""" SHA1 hash-generator
	"""
	__pycrypto_cls__ = SHA1
	__generator_name__ = 'SHA1'


class WSHA224(WSHAFamily):
	""" SHA224 hash-generator
	"""
	__pycrypto_cls__ = SHA224
	__generator_name__ = 'SHA224'


class WSHA256(WSHAFamily):
	""" SHA256 hash-generator
	"""
	__pycrypto_cls__ = SHA256
	__generator_name__ = 'SHA256'


class WSHA384(WSHAFamily):
	""" SHA384 hash-generator
	"""
	__pycrypto_cls__ = SHA384
	__generator_name__ = 'SHA384'


class WSHA512(WSHAFamily):
	""" SHA512 hash-generator
	"""
	__pycrypto_cls__ = SHA512
	__generator_name__ = 'SHA512'


class WMD5(WPyCryptoHashAdapter):
	""" MD5 hash-generator
	"""
	__pycrypto_cls__ = MD5
	__generator_name__ = 'MD5'


class WHash:
	""" Class that aggregates different hash-generators. This class is should be used if there is a need to address
	digest generator by its name. As a result - generator (:class:`.WHashGeneratorProto`) is returned.
	"""

	__hash_map__ = {x.generator_name(): x for x in (WSHA1, WSHA224, WSHA256, WSHA384, WSHA512, WMD5)}
	""" Available hash generators map 'hash function name' - :class:`.WHashGeneratorProto`
	"""

	@staticmethod
	@verify_type(name=str)
	def generator(name):
		""" Return generator by its name

		:param name: name of hash-generator

		:return: WHashGeneratorProto class
		"""
		name = name.upper()
		if name not in WHash.__hash_map__.keys():
			raise ValueError('Hash generator "%s" not available' % name)
		return WHash.__hash_map__[name]

	@staticmethod
	def generator_by_digest(family, digest_size):
		""" Return generator by hash generator family name and digest size

		:param family: name of hash-generator family

		:return: WHashGeneratorProto class
		"""
		for generator_name in WHash.available_generators(family=family):
			generator = WHash.generator(generator_name)
			if generator.generator_digest_size() == digest_size:
				return generator
		raise ValueError('Hash generator is not available')

	@staticmethod
	@verify_type(family=(str, None), name=(str, None))
	def available_generators(family=None, name=None):
		""" Return names of available generators

		:param family: name of hash-generator family to select
		:param name: name of hash-generator to select (parameter may be used for availability check)

		:return: tuple of str
		"""
		generators = WHash.__hash_map__.values()

		if family is not None:
			family = family.upper()
			generators = filter(lambda x: x.generator_family() == family, generators)

		if name is not None:
			name = name.upper()
			generators = filter(lambda x: x.generator_name() == name, generators)

		return tuple([x.generator_name() for x in generators])

	@staticmethod
	@verify_type(family=(str, None), name=(str, None))
	def available_digests(family=None, name=None):
		""" Return names of available generators

		:param family: name of hash-generator family to select
		:param name: name of hash-generator to select

		:return: set of int
		"""
		generators = WHash.available_generators(family=family, name=name)
		return set([WHash.generator(x).generator_digest_size() for x in generators])



# -*- coding: utf-8 -*-

import pytest

from wasp_general.uri import WURI, WSchemeSpecification, WSchemeHandler, WSchemeCollection


def test_abstract():
	pytest.raises(TypeError, WSchemeHandler)
	pytest.raises(NotImplementedError, WSchemeHandler.create_handler, WURI())
	pytest.raises(NotImplementedError, WSchemeHandler.scheme_specification)


class TestWURI:

	def test(self):
		uri = WURI()
		assert(isinstance(uri, WURI) is True)

		assert(uri.scheme() is None)
		assert(uri.username() is None)
		assert(uri.password() is None)
		assert(uri.hostname() is None)
		assert(uri.port() is None)
		assert(uri.path() is None)
		assert(uri.query() is None)
		assert(uri.fragment() is None)
		assert(str(uri) == '')
		assert(
			[x for x in uri] == [
				(WURI.Component.scheme, None),
				(WURI.Component.username, None),
				(WURI.Component.password, None),
				(WURI.Component.hostname, None),
				(WURI.Component.port, None),
				(WURI.Component.path, None),
				(WURI.Component.query, None),
				(WURI.Component.fragment, None)
			]
		)

		uri = WURI.parse(str(uri))
		assert(uri.scheme() is None)
		assert(uri.username() is None)
		assert(uri.password() is None)
		assert(uri.hostname() is None)
		assert(uri.port() is None)
		assert(uri.path() is None)
		assert(uri.query() is None)
		assert(uri.fragment() is None)

		uri = WURI(scheme='proto', hostname='host1', path='/foo')

		assert(uri.scheme() == 'proto')
		assert(uri.username() is None)
		assert(uri.password() is None)
		assert(uri.hostname() == 'host1')
		assert(uri.port() is None)
		assert(uri.path() == '/foo')
		assert(uri.query() is None)
		assert(uri.fragment() is None)
		assert(str(uri) == 'proto://host1/foo')
		assert(
			[x for x in uri] == [
				(WURI.Component.scheme, 'proto'),
				(WURI.Component.username, None),
				(WURI.Component.password, None),
				(WURI.Component.hostname, 'host1'),
				(WURI.Component.port, None),
				(WURI.Component.path, '/foo'),
				(WURI.Component.query, None),
				(WURI.Component.fragment, None)
			]
		)

		uri = WURI.parse(str(uri))
		assert(uri.scheme() == 'proto')
		assert(uri.username() is None)
		assert(uri.password() is None)
		assert(uri.hostname() == 'host1')
		assert(uri.port() is None)
		assert(uri.path() == '/foo')
		assert(uri.query() is None)
		assert(uri.fragment() is None)

		uri = WURI(
			scheme='proto', username='local_user', password='secret', hostname='host1', port=40,
			path='/foo', query='q=10;p=2', fragment='section1'
		)

		assert(uri.scheme() == 'proto')
		assert(uri.username() == 'local_user')
		assert(uri.password() == 'secret')
		assert(uri.hostname() == 'host1')
		assert(uri.port() == 40)
		assert(uri.path() == '/foo')
		assert(uri.query() == 'q=10;p=2')
		assert(uri.fragment() == 'section1')
		assert(str(uri) == 'proto://local_user:secret@host1:40/foo?q=10;p=2#section1')
		assert(
			[x for x in uri] == [
				(WURI.Component.scheme, 'proto'),
				(WURI.Component.username, 'local_user'),
				(WURI.Component.password, 'secret'),
				(WURI.Component.hostname, 'host1'),
				(WURI.Component.port, 40),
				(WURI.Component.path, '/foo'),
				(WURI.Component.query, 'q=10;p=2'),
				(WURI.Component.fragment, 'section1')
			]
		)

		uri = WURI.parse(str(uri))
		assert(uri.scheme() == 'proto')
		assert(uri.username() == 'local_user')
		assert(uri.password() == 'secret')
		assert(uri.hostname() == 'host1')
		assert(uri.port() == 40)
		assert(uri.path() == '/foo')
		assert(uri.query() == 'q=10;p=2')
		assert(uri.fragment() == 'section1')

		uri = WURI(scheme='proto', path='/foo')
		assert(str(uri) == 'proto:///foo')

		uri = WURI.parse(str(uri))
		assert(uri.scheme() == 'proto')
		assert(uri.username() is None)
		assert(uri.password() is None)
		assert(uri.hostname() is None)
		assert(uri.port() is None)
		assert(uri.path() == '/foo')
		assert(uri.query() is None)
		assert(uri.fragment() is None)

		pytest.raises(AttributeError, "uri.zzz")


class TestWSchemeSpecification:

	def test(self):
		scheme_spec = WSchemeSpecification('proto')
		assert(isinstance(scheme_spec, WSchemeSpecification) is True)
		assert(scheme_spec.scheme_name() == 'proto')

		assert(
			[scheme_spec.descriptor(x) for x in WURI.Component] == [
				WSchemeSpecification.ComponentDescriptor.required,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported
			]
		)

		assert(
			[x for x in scheme_spec] == [
				(WURI.Component.scheme, WSchemeSpecification.ComponentDescriptor.required),
				(WURI.Component.username, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.password, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.hostname, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.port, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.path, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.query, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.fragment, WSchemeSpecification.ComponentDescriptor.unsupported)
			]
		)

		assert(scheme_spec.is_compatible(WURI.parse('proto:')) is True)
		assert(scheme_spec.is_compatible(WURI.parse('proto://host')) is False)

		scheme_spec = WSchemeSpecification(
			'proto',
			hostname=WSchemeSpecification.ComponentDescriptor.required,
			port=WSchemeSpecification.ComponentDescriptor.optional
		)

		assert(
			[scheme_spec.descriptor(x) for x in WURI.Component] == [
				WSchemeSpecification.ComponentDescriptor.required,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.required,
				WSchemeSpecification.ComponentDescriptor.optional,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported,
				WSchemeSpecification.ComponentDescriptor.unsupported
			]
		)

		assert(
			[x for x in scheme_spec] == [
				(WURI.Component.scheme, WSchemeSpecification.ComponentDescriptor.required),
				(WURI.Component.username, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.password, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.hostname, WSchemeSpecification.ComponentDescriptor.required),
				(WURI.Component.port, WSchemeSpecification.ComponentDescriptor.optional),
				(WURI.Component.path, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.query, WSchemeSpecification.ComponentDescriptor.unsupported),
				(WURI.Component.fragment, WSchemeSpecification.ComponentDescriptor.unsupported)
			]
		)

		assert(scheme_spec.is_compatible(WURI.parse('proto:')) is False)
		assert(scheme_spec.is_compatible(WURI.parse('proto://host')) is True)
		assert(scheme_spec.is_compatible(WURI.parse('proto://host:30')) is True)
		assert(scheme_spec.is_compatible(WURI.parse('proto:///')) is False)

		pytest.raises(
			TypeError, WSchemeSpecification,
			'proto', scheme=WSchemeSpecification.ComponentDescriptor.required
		)

		pytest.raises(TypeError, WSchemeSpecification, 'proto', hostname='optional')


class TestWSchemeCollection:

	# noinspection PyAbstractClass
	class Handler(WSchemeHandler):

		def __init__(self, uri):
			WSchemeHandler.__init__(self)
			self.uri = uri

		@classmethod
		def create_handler(cls, uri):
			return cls(uri)

	class HandlerFoo(Handler):

		@classmethod
		def scheme_specification(cls):
			return WSchemeSpecification('foo', path=WSchemeSpecification.ComponentDescriptor.required)

	class HandlerBar(Handler):

		@classmethod
		def scheme_specification(cls):
			return WSchemeSpecification(
				'bar',
				path=WSchemeSpecification.ComponentDescriptor.required,
				query=WSchemeSpecification.ComponentDescriptor.optional
			)

	def test_exceptions(self):
		e = WSchemeCollection.NoHandlerFound(WURI())
		assert(isinstance(e, WSchemeCollection.NoHandlerFound) is True)
		assert(isinstance(e, Exception) is True)

		e = WSchemeCollection.SchemeIncompatible(WURI())
		assert(isinstance(e, WSchemeCollection.SchemeIncompatible) is True)
		assert(isinstance(e, Exception) is True)

	def test(self):
		uri1 = WURI.parse('/path')
		uri2 = WURI.parse('foo:///path')
		uri3 = WURI.parse('bar:///path')
		uri4 = WURI.parse('bar:///path?test=foo')
		uri5 = WURI.parse('foo:///path?test=foo')

		collection = WSchemeCollection()
		assert(isinstance(collection, WSchemeCollection) is True)
		assert(collection.handler() is None)
		assert(collection.handler(scheme_name='foo') is None)
		assert(collection.handler(scheme_name='bar') is None)

		pytest.raises(WSchemeCollection.NoHandlerFound, collection.open, uri1)
		pytest.raises(WSchemeCollection.NoHandlerFound, collection.open, uri2)
		pytest.raises(WSchemeCollection.NoHandlerFound, collection.open, uri3)
		pytest.raises(WSchemeCollection.NoHandlerFound, collection.open, uri4)
		pytest.raises(WSchemeCollection.NoHandlerFound, collection.open, uri5)

		collection = WSchemeCollection(
			TestWSchemeCollection.HandlerFoo, default_handler_cls=TestWSchemeCollection.HandlerBar
		)
		assert(isinstance(collection, WSchemeCollection) is True)
		assert(collection.handler() == TestWSchemeCollection.HandlerBar)
		assert(collection.handler(scheme_name='foo') == TestWSchemeCollection.HandlerFoo)
		assert(collection.handler(scheme_name='bar') is None)

		assert(isinstance(collection.open(uri1), TestWSchemeCollection.HandlerBar) is True)
		assert(isinstance(collection.open(uri2), TestWSchemeCollection.HandlerFoo) is True)
		pytest.raises(WSchemeCollection.NoHandlerFound, collection.open, uri3)
		pytest.raises(WSchemeCollection.NoHandlerFound, collection.open, uri4)
		pytest.raises(WSchemeCollection.SchemeIncompatible, collection.open, uri5)

		collection.add(TestWSchemeCollection.HandlerBar)
		assert(isinstance(collection.open(uri1), TestWSchemeCollection.HandlerBar) is True)
		assert(isinstance(collection.open(uri2), TestWSchemeCollection.HandlerFoo) is True)
		assert(isinstance(collection.open(uri3), TestWSchemeCollection.HandlerBar) is True)
		assert(isinstance(collection.open(uri4), TestWSchemeCollection.HandlerBar) is True)
		pytest.raises(WSchemeCollection.SchemeIncompatible, collection.open, uri5)



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.messenger.proto import WMessengerOnionSessionFlowProto, WMessengerOnionProto
from wasp_general.network.messenger.proto import WMessengerOnionLayerProto
from wasp_general.network.messenger.envelope import WMessengerEnvelope, WMessengerTextEnvelope, WMessengerBytesEnvelope

from wasp_general.network.messenger.session import WMessengerOnionSessionFlow, WMessengerOnionSessionFlowSequence
from wasp_general.network.messenger.session import WMessengerOnionSessionFlexFlow, WMessengerOnionSession


def test_abstract():

	pytest.raises(TypeError, WMessengerOnionSessionFlexFlow.MessageComparator)
	envelope = WMessengerEnvelope(None)
	pytest.raises(NotImplementedError, WMessengerOnionSessionFlexFlow.MessageComparator.match, None, envelope)


class TestWMessengerOnionSessionFlow:

	@staticmethod
	def expand(iterator=None, envelope_data=None):
		result = []
		envelope = WMessengerEnvelope(envelope_data)
		if iterator is not None:
			result.append(
				WMessengerOnionSessionFlow.IteratorInfo(iterator.layer_name(), **iterator.layer_args())
			)
			result.extend(TestWMessengerOnionSessionFlow.expand(iterator.next(envelope)))
		return result

	def test(self):
		i = WMessengerOnionSessionFlowProto.Iterator('layer')
		sf = WMessengerOnionSessionFlow(i)

		assert(isinstance(sf, WMessengerOnionSessionFlow) is True)
		assert(isinstance(sf, WMessengerOnionSessionFlowProto) is True)
		assert(sf.iterator(WMessengerEnvelope(None)) == i)

		info = [
			WMessengerOnionSessionFlowProto.IteratorInfo('layer1', **{'a': 1, 'b': 'z'}),
			WMessengerOnionSessionFlowProto.IteratorInfo('layer2', **{'a': 2, 'b': 'zz'}),
			WMessengerOnionSessionFlowProto.IteratorInfo('layer1', **{'a': 3, 'b': 'z'})
		]

		i = WMessengerOnionSessionFlow.sequence(*info)
		assert(isinstance(i, WMessengerOnionSessionFlowProto.Iterator) is True)
		result = TestWMessengerOnionSessionFlow.expand(i)
		assert([(x.layer_name(), x.layer_args()) for x in result] == [
			('layer1', {'a': 1, 'b': 'z'}),
			('layer2', {'a': 2, 'b': 'zz'}),
			('layer1', {'a': 3, 'b': 'z'})
		])

		i = WMessengerOnionSessionFlow.sequence()
		assert(i is None)

		sf = WMessengerOnionSessionFlow.sequence_flow(*info)
		assert(isinstance(sf, WMessengerOnionSessionFlowProto) is True)
		i = sf.iterator(WMessengerEnvelope(None))
		result = TestWMessengerOnionSessionFlow.expand(i)
		assert([(x.layer_name(), x.layer_args()) for x in result] == [
			('layer1', {'a': 1, 'b': 'z'}),
			('layer2', {'a': 2, 'b': 'zz'}),
			('layer1', {'a': 3, 'b': 'z'})
		])


class TestWMessengerOnionSessionFlowSequence:

	def test_iterator(self):
		i = WMessengerOnionSessionFlowSequence.FlowSequenceIterator(
			WMessengerOnionSessionFlow.IteratorInfo('layer', **{'z': 1})
		)
		assert(isinstance(i, WMessengerOnionSessionFlowSequence.FlowSequenceIterator) is True)
		assert(isinstance(i, WMessengerOnionSessionFlow.Iterator) is True)
		assert(i.layer_name() == 'layer')
		assert(i.layer_args() == {'z': 1})
		result = TestWMessengerOnionSessionFlow.expand(i)
		assert(len(result) == 1)

		i11 = WMessengerOnionSessionFlow.IteratorInfo('layer1-1')
		i12 = WMessengerOnionSessionFlow.IteratorInfo('layer1-2', **{'q': 0})
		f1 = WMessengerOnionSessionFlow.sequence_flow(i11, i12)

		i21 = WMessengerOnionSessionFlow.IteratorInfo('layer2-1')
		i22 = WMessengerOnionSessionFlow.IteratorInfo('layer2-2')
		i23 = WMessengerOnionSessionFlow.IteratorInfo('layer2-3', **{'a': 'w'})
		f2 = WMessengerOnionSessionFlow.sequence_flow(i21, i22, i23)

		f3 = WMessengerOnionSessionFlow.sequence_flow(i11)
		f3.iterator = lambda x: None

		i = WMessengerOnionSessionFlowSequence.FlowSequenceIterator(
			WMessengerOnionSessionFlow.IteratorInfo('layer', **{'0': 0}), f1, f3, f2
		)
		result = TestWMessengerOnionSessionFlow.expand(i)
		assert([(x.layer_name(), x.layer_args()) for x in result] == [
			('layer', {'0': 0}),
			('layer1-1', {}),
			('layer1-2', {'q': 0}),
			('layer2-1', {}),
			('layer2-2', {}),
			('layer2-3', {'a': 'w'}),
		])

	def test(self):
		i11 = WMessengerOnionSessionFlow.IteratorInfo('layer1')
		i12 = WMessengerOnionSessionFlow.IteratorInfo('layer2')
		f1 = WMessengerOnionSessionFlow.sequence_flow(i11, i12)

		i21 = WMessengerOnionSessionFlow.IteratorInfo('layer3')
		i22 = WMessengerOnionSessionFlow.IteratorInfo('layer4')
		i23 = WMessengerOnionSessionFlow.IteratorInfo('layer5')
		f2 = WMessengerOnionSessionFlow.sequence_flow(i21, i22, i23)

		sf = WMessengerOnionSessionFlowSequence()
		assert(isinstance(sf, WMessengerOnionSessionFlowSequence) is True)
		assert(isinstance(sf, WMessengerOnionSessionFlowProto) is True)
		assert(sf.iterator(WMessengerEnvelope(None)) is None)

		sf = WMessengerOnionSessionFlowSequence(f1, f2)
		result = sf.iterator(WMessengerEnvelope(None))
		assert(isinstance(result, WMessengerOnionSessionFlowSequence.FlowSequenceIterator) is True)
		result = TestWMessengerOnionSessionFlow.expand(result)
		assert([(x.layer_name(), x.layer_args()) for x in result] == [
			('layer1', {}),
			('layer2', {}),
			('layer3', {}),
			('layer4', {}),
			('layer5', {})
		])


class TestWMessengerOnionSessionFlexFlow:
	class CMP(WMessengerOnionSessionFlexFlow.MessageComparator):
		def match(self, envelope):
			return len(envelope.message()) % 2 == 0

	class Dummy(WMessengerOnionSessionFlexFlow.MessageComparator):

		def __init__(self, value):
			self.__value = value

		def match(self, envelope):
			return self.__value

	def test_comparator_pair(self):
		cmp = TestWMessengerOnionSessionFlexFlow.CMP()

		assert(isinstance(cmp, WMessengerOnionSessionFlexFlow.MessageComparator) is True)

		f = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlow.IteratorInfo('layer'))
		pytest.raises(TypeError, WMessengerOnionSessionFlexFlow.FlowComparatorPair, 1, f)
		p = WMessengerOnionSessionFlexFlow.FlowComparatorPair(cmp, f)
		assert(p.comparator() == cmp)
		assert(p.flow() == f)

		re_cmp = WMessengerOnionSessionFlexFlow.ReComparator('^zxc.+11$')
		assert(re_cmp.match(WMessengerTextEnvelope('zxc11')) is not True)
		assert(re_cmp.match(WMessengerTextEnvelope('zxc_11')) is True)

		re_cmp = WMessengerOnionSessionFlexFlow.ReComparator(b'zxc_11')
		assert(re_cmp.match(WMessengerBytesEnvelope(b'zxc_11')) is True)
		assert(re_cmp.match(WMessengerBytesEnvelope(b'zxc')) is not True)

	def test(self):

		cmp = TestWMessengerOnionSessionFlexFlow.CMP()
		f1 = WMessengerOnionSessionFlow(WMessengerOnionSessionFlow.Iterator('layer0'))
		p1 = WMessengerOnionSessionFlexFlow.FlowComparatorPair(cmp, f1)

		dummy_true_cmp = TestWMessengerOnionSessionFlexFlow.Dummy(True)
		f2 = WMessengerOnionSessionFlow.sequence_flow(
			WMessengerOnionSessionFlowProto.IteratorInfo('layer1'),
			WMessengerOnionSessionFlowProto.IteratorInfo('layer2')
		)
		p2 = WMessengerOnionSessionFlexFlow.FlowComparatorPair(dummy_true_cmp, f2)

		ff = WMessengerOnionSessionFlexFlow(p1, p2)
		expanded_result = TestWMessengerOnionSessionFlow.expand(ff.iterator(WMessengerTextEnvelope('q')))
		assert([x.layer_name() for x in expanded_result] == ['layer1', 'layer2'])
		expanded_result = TestWMessengerOnionSessionFlow.expand(ff.iterator(WMessengerTextEnvelope('qq')))
		assert([x.layer_name() for x in expanded_result] == ['layer0'])

		f3 = WMessengerOnionSessionFlow.sequence_flow(WMessengerOnionSessionFlowProto.IteratorInfo('layer4'))

		ff = WMessengerOnionSessionFlexFlow(p1, default_flow=f3)
		expanded_result = TestWMessengerOnionSessionFlow.expand(ff.iterator(WMessengerTextEnvelope('q')))
		assert([x.layer_name() for x in expanded_result] == ['layer4'])
		expanded_result = TestWMessengerOnionSessionFlow.expand(ff.iterator(WMessengerTextEnvelope('qq')))
		assert([x.layer_name() for x in expanded_result] == ['layer0'])

		pytest.raises(TypeError, WMessengerOnionSessionFlexFlow, 4)


class TestWMessengerOnionSession:

	def test(self):

		class Onion(WMessengerOnionProto):

			def __init__(self):
				self.storage = {}

			def layers_names(self):
				return self.storage.keys()

			def layer(self, layer_name):
				return self.storage[layer_name]

		class Layer(WMessengerOnionLayerProto):

			def process(self, envelope, session, **kwargs):
				return WMessengerEnvelope(envelope.message() + '<' + self.name() + '>::')

		class CustomSessionFlow(WMessengerOnionSessionFlowProto):

			def __init__(self):
				WMessengerOnionSessionFlowProto.__init__(self)

				self.__items = WMessengerOnionSessionFlow.Iterator(
					'layer1', WMessengerOnionSessionFlow.Iterator(
						'layer2', WMessengerOnionSessionFlow.Iterator(
							'layer1', WMessengerOnionSessionFlow.Iterator(
								'l3'
							)
						)
					)
				)

			def iterator(self, envelope):
				return self.__items

		onion = Onion()
		session_flow = CustomSessionFlow()
		onion_session = WMessengerOnionSession(onion, session_flow)

		layers = [Layer('layer'), Layer('layer2'), Layer('last_layer'), Layer('l3'), Layer('layer1')]
		for l in layers:
			onion.storage[l.name()] = l

		assert(onion_session.onion() == onion)
		assert(onion_session.session_flow() == session_flow)
		result = onion_session.process(WMessengerEnvelope('msg'))
		assert(isinstance(result, WMessengerEnvelope) is True)
		assert(result.message() == 'msg<layer1>::<layer2>::<layer1>::<l3>::')
